# sage_setup: distribution = sagemath-combinat
r"""
Suffix tries and suffix trees
"""
# ****************************************************************************
#       Copyright (C) 2008 Franco Saliola <saliola@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from itertools import chain

from sage.structure.sage_object import SageObject
from sage.sets.set import Set
from sage.combinat.words.words import Words
from sage.combinat.words.word import Word
from sage.misc.lazy_import import lazy_import
from sage.rings.integer import Integer

lazy_import('sage.graphs.digraph', 'DiGraph')


# ------------
# Suffix Tries
# ------------


class SuffixTrie(SageObject):
    def __init__(self, word):
        r"""
        Construct the suffix trie of the word w.

        The suffix trie of a finite word w is a data structure representing
        the factors of w. It is a tree whose edges are labelled with
        letters of w, and whose leafs correspond to suffixes of w.

        This is a straightforward implementation of Algorithm 1 from
        [Ukko1995]_.  It constructs the suffix trie of w[:i] from that
        of w[:i-1].

        A suffix trie is modelled as a deterministic finite-state automaton
        together with the suffix_link map. The set of states corresponds to
        factors of the word (below we write x' for the state corresponding
        to x); these are always 0, 1, .... The state 0 is the initial
        state, and it corresponds to the empty word.  For the purposes of
        the algorithm, there is also an auxiliary state -1. The transition
        function t is defined as::

                t(-1,a) = 0 for all letters a; and
                t(x',a) = y' for all x',y' \in Q such that y = xa,

        and the suffix link function is defined as::

                suffix_link(0) = -1;
                suffix_link(x') = y', if x = ay for some letter a.

        REFERENCES:

        - [Ukko1995]_

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: t = SuffixTrie(w); t
            Suffix Trie of the word: cacao

        ::

            sage: e = Words("ab")()
            sage: t = SuffixTrie(e); t
            Suffix Trie of the word:
            sage: t.process_letter("a"); t
            Suffix Trie of the word: a
            sage: t.process_letter("b"); t
            Suffix Trie of the word: ab
            sage: t.process_letter("a"); t
            Suffix Trie of the word: aba

        TESTS::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: s = SuffixTrie(w)
            sage: loads(dumps(s))
            Suffix Trie of the word: cacao
        """
        # Create the suffix trie for the empty word.
        self._active_state = 0
        self._transition_function = {}
        self._suffix_link = [-1]
        self._alphabet = word.parent().alphabet()

        # Process each letter, in order.
        W = word.parent()
        for letter in word:
            self._process_letter(W([letter]))

    def _process_letter(self, letter):
        r"""
        Process a letter. That is, modify the current suffix trie producing
        the suffix trie for ``self.word() + letter``.

        .. NOTE::

           ``letter`` must occur within the alphabet of the word.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: t = SuffixTrie(Word("ababba"))
            sage: t._process_letter(Words("ab")("b")); t
            Suffix Trie of the word: ababbab
        """
        r = self._active_state
        old_s = None
        # While r is not the auxiliary vertex, or
        # there is not transition from r along letter, ...
        while r != -1 and (r, letter) not in self._transition_function:
            # adjoin a new state s
            s = len(self._suffix_link)
            self._suffix_link.append(None)
            # create a transition from r to s along letter
            self._transition_function[(r, letter)] = s
            if r != self._active_state:
                # update the suffix link
                self._suffix_link[old_s] = s
            old_s = s
            r = self._suffix_link[r]
        # update the suffix link for the last visited state
        if r == -1:
            self._suffix_link[old_s] = 0
        else:
            self._suffix_link[old_s] = self._transition_function[(r, letter)]
        # update the active state
        self._active_state = \
            self._transition_function[(self._active_state, letter)]

    def process_letter(self, letter):
        r"""
        Modify ``self`` to produce the suffix trie for ``self.word() +
        letter``.

        .. NOTE::

           ``letter`` must occur within the alphabet of the word.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("ab")("ababba")
            sage: t = SuffixTrie(w); t
            Suffix Trie of the word: ababba
            sage: t.process_letter("a"); t
            Suffix Trie of the word: ababbaa

        TESTS::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: t = SuffixTrie(w); t
            Suffix Trie of the word: cacao
            sage: t.process_letter("d")
            Traceback (most recent call last):
            ...
            ValueError: d not in alphabet
        """
        # Make certain that letter is a word containing one letter.
        letter = Words(self._alphabet)([letter])
        self._process_letter(letter)

    #####
    # The following are not necessary for constructing the suffix trie (just
    # the __init__ and process_letter are needed). They just add additional
    # functionality to the class.
    #####

    def _repr_(self) -> str:
        """
        TESTS::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: SuffixTrie(Word("abcba"))._repr_()
            'Suffix Trie of the word: abcba'
        """
        return 'Suffix Trie of the %s' % repr(self.word())

    def node_to_word(self, state=0):
        r"""
        Return the word obtained by reading the edge labels from 0 to
        ``state``.

        INPUT:

        - ``state`` -- (default: 0) a state

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("abc")("abcba")
            sage: t = SuffixTrie(w)
            sage: t.node_to_word(10)
            word: abcba
            sage: t.node_to_word(7)
            word: abcb
        """
        if state == 0:
            return Words(self._alphabet)()
        # We first invert the transition function
        tf_inv = {b: a for a, b in self._transition_function.items()}

        # Starting from the active state,
        # read labels along the unique path to the root.
        u, letter = tf_inv[state]
        w = letter
        s = u
        while s != 0:
            u, letter = tf_inv[s]
            w = letter * w
            s = u
        return w

    def word(self):
        r"""
        Return the word whose suffix tree this is.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("abc")("abcba")
            sage: t = SuffixTrie(w)
            sage: t.word()
            word: abcba
            sage: t.word() == w
            True
        """
        return self.node_to_word(self._active_state)

    def __eq__(self, other) -> bool:
        r"""
        If ``self`` and ``other`` have the same transition function,
        the same suffix link, and the same word, then they are equal.

        TESTS::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: SuffixTrie(Word("cacao")) == SuffixTrie(Word("ababc"))
            False
            sage: W = Words("cao")
            sage: s = SuffixTrie(W("cacao"))
            sage: t = SuffixTrie(W())
            sage: t.process_letter("c")
            sage: t.process_letter("a")
            sage: t.process_letter("c")
            sage: t.process_letter("a")
            sage: t.process_letter("o")
            sage: t == s
            True
        """
        if not isinstance(other, SuffixTrie):
            return False
        return self._transition_function == other._transition_function \
            and self._suffix_link == other._suffix_link \
            and self.word() == other.word()

    def transition_function(self, node, word):
        r"""
        Return the state reached by beginning at ``node`` and following the
        arrows in the transition graph labelled by the letters of ``word``.

        INPUT:

        - ``node`` -- a node
        - ``word`` -- a word

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words([0,1])([0,1,0,1,1])
            sage: t = SuffixTrie(w)
            sage: all(t.transition_function(u, letter) == v
            ....:     for (u, letter), v in t._transition_function.items())
            True
        """
        if node == -1:
            return self.transition_function(0, word[1:])
        if word.is_empty():
            return 0
        if word.length() == 1:
            return self._transition_function[(node, word)]
        return self.transition_function(
            self._transition_function[(node, word[0:1])], word[1:])

    def states(self):
        r"""
        Return the states of the automaton defined by the suffix trie.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words([0,1])([0,1,1])
            sage: t = SuffixTrie(w)
            sage: t.states()
            [0, 1, 2, 3, 4]

        ::

            sage: u = Words("aco")("cacao")
            sage: s = SuffixTrie(u)
            sage: s.states()
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        """
        return list(range(len(self._transition_function)))

    def suffix_link(self, state):
        r"""
        Evaluate the suffix link map of the suffix trie on ``state``.

        Note that the suffix link map is not defined on -1.

        INPUT:

        - ``state`` -- a state

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: t = SuffixTrie(w)
            sage: list(map(t.suffix_link, range(13)))
            [-1, 0, 3, 0, 5, 1, 7, 2, 9, 10, 11, 12, 0]
            sage: t.suffix_link(0)
            -1

        TESTS::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: t = SuffixTrie(w)
            sage: t.suffix_link([1])
            Traceback (most recent call last):
            ...
            TypeError: [1] is not an integer
            sage: t.suffix_link(-1)
            Traceback (most recent call last):
            ...
            TypeError: suffix link is not defined for -1
            sage: t.suffix_link(17)
            Traceback (most recent call last):
            ...
            TypeError: 17 is not a state
        """
        if not isinstance(state, (int, Integer)):
            raise TypeError("%s is not an integer" % state)
        if state == -1:
            raise TypeError("suffix link is not defined for -1")
        if state not in range(len(self._suffix_link)):
            raise TypeError("%s is not a state" % state)
        return self._suffix_link[state]

    def active_state(self):
        r"""
        Return the active state of the suffix trie.

        This is the state corresponding to the word as a suffix of
        itself.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: t = SuffixTrie(w)
            sage: t.active_state()
            8

        ::

            sage: u = Words([0,1])([0,1,1,0,1,0,0,1])
            sage: s = SuffixTrie(u)
            sage: s.active_state()
            22
        """
        return self._active_state

    def final_states(self):
        r"""
        Return the set of final states of the suffix trie.

        These are the states corresponding to the suffixes of
        ``self.word()``. They are obtained be repeatedly following the
        suffix link from the active state until we reach 0.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: t = SuffixTrie(w)
            sage: t.final_states() == Set([8, 9, 10, 11, 12, 0])
            True
        """
        s = self._active_state
        F = [s]
        while s != 0:
            s = self._suffix_link[s]
            F.append(s)
        return Set(F)

    def has_suffix(self, word) -> bool:
        r"""
        Return ``True`` if and only if ``word`` is a suffix of ``self.word()``.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cacao")
            sage: t = SuffixTrie(w)
            sage: [t.has_suffix(w[i:]) for i in range(w.length()+1)]
            [True, True, True, True, True, True]
            sage: [t.has_suffix(w[:i]) for i in range(w.length()+1)]
            [True, False, False, False, False, True]
        """
        # Find the state corresponding to word, and
        # check to see if s is a final state.
        s = self.transition_function(0, word)
        q = self._active_state
        if q == s:
            return True

        while q != 0:
            q = self._suffix_link[q]
            if q == s:
                return True
        return False

    def to_digraph(self):
        r"""
        Return a ``DiGraph`` object of the transition graph of the suffix
        trie.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cac")
            sage: t = SuffixTrie(w)
            sage: d = t.to_digraph(); d                                                 # needs sage.graphs
            Digraph on 6 vertices
            sage: d.adjacency_matrix()                                                  # needs sage.graphs sage.modules
            [0 1 0 1 0 0]
            [0 0 1 0 0 0]
            [0 0 0 0 1 0]
            [0 0 0 0 0 1]
            [0 0 0 0 0 0]
            [0 0 0 0 0 0]
        """
        dag = {}
        for (u, letter), v in self._transition_function.items():
            dag.setdefault(u, {})[v] = letter
        return DiGraph(dag)

    def plot(self, layout='tree', tree_root=0, tree_orientation='up',
             vertex_colors=None, edge_labels=True, *args, **kwds):
        r"""
        Return a Graphics object corresponding to the transition graph of
        the suffix trie.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: SuffixTrie(Word("cacao")).plot()                                      # needs sage.plot
            Graphics object consisting of 38 graphics primitives

        TESTS::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: type(SuffixTrie(Word("cacao")).plot())                                # needs sage.plot
            <class 'sage.plot.graphics.Graphics'>
        """
        tree = self.to_digraph()
        for u, v, label in tree.edge_iterator():
            tree.set_edge_label(u, v, label.string_rep())
        if vertex_colors is None:
            suffix_nodes = self.final_states()
            non_suffix_nodes = list(set(self.states()) - set(suffix_nodes))
            vertex_colors = {'#fec7b8': suffix_nodes, '#ffffff': non_suffix_nodes}
        return tree.plot(layout=layout, tree_root=tree_root,
                         tree_orientation=tree_orientation,
                         vertex_colors=vertex_colors, edge_labels=edge_labels,
                         *args, **kwds)

    def show(self, *args, **kwds):
        r"""
        Display the output of :meth:`plot`.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import SuffixTrie
            sage: w = Words("cao")("cac")
            sage: t = SuffixTrie(w)
            sage: t.show()                                                              # needs sage.plot
        """
        self.plot(*args, **kwds).show()


# ------------
# Suffix Trees
# ------------


class ImplicitSuffixTree(SageObject):
    def __init__(self, word):
        r"""
        Construct the implicit suffix tree of a word w.

        The suffix tree of a word w is a compactification of the suffix
        trie for w. The compactification removes all nodes that have
        exactly one incoming edge and exactly one outgoing edge. It
        consists of two components: a tree and a word. Thus, instead of
        labelling the edges by factors of w, we can labelled them by
        indices of the occurrence of the factors in w.

        The following is a straightforward implementation of Ukkonen's
        on-line algorithm for constructing the
        implicit suffix tree [Ukko1995]_.  It constructs the suffix tree for
        w[:i] from that of w[:i-1].

        GENERAL IDEA. The suffix tree of w[:i+1] can be obtained from that
        of w[:i] by visiting each node corresponding to a suffix of w[:i]
        and modifying the tree by applying one of two rules (either append
        a new node to the tree, or split an edge into two). The "active
        state" is the node where the algorithm begins and the "suffix link"
        carries us to the next node that needs to be dealt with.

        TREE. The tree is modelled as an automaton, which is stored as a
        dictionary of dictionaries: it is keyed by the nodes of the tree,
        and the corresponding dictionary is keyed by pairs `(i,j)` of
        integers representing the word w[i-1:j]. This makes it faster to
        look up a particular transition beginning at a specific node.

        STATES/NODES. The states will always be -1, 0, 1, ..., n. The state
        -1 is special and is only used for the purposes of the algorithm.
        All transitions map -1 to 0, so this information is not explicitly
        stored in the transition function.

        EXPLICIT/IMPLICIT NODES. By definition, some of the nodes will not
        be states, but merely locations along an edge; these are called
        implicit nodes. A node r (implicit or explicit) is referenced as a
        pair (s,(k,p)) where s is an ancestor of r and w[k-1:p] is the word
        read by transitioning from s to r in the suffix trie. A reference
        pair is canonical if s is the closest ancestor of r.

        SUFFIX LINK. The algorithm makes use of a map from (some) nodes to
        other nodes, called the suffix link. This is stored as a
        dictionary.

        ACTIVE STATE. We store as ._active_state the active state of the
        tree, the state where the algorithm will begin when processing the
        next letter.

        RUNNING TIME. The running time and storage space of the algorithm
        is linear in the length of the word w (whereas for a suffix tree it
        is quadratic).

        REFERENCES:

        - [Ukko1995]_

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words("aco")("cacao")
            sage: t = ImplicitSuffixTree(w); t
            Implicit Suffix Tree of the word: cacao
            sage: ababb = Words([0,1])([0,1,0,1,1])
            sage: s = ImplicitSuffixTree(ababb); s
            Implicit Suffix Tree of the word: 01011

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words("cao")("cacao")
            sage: s = ImplicitSuffixTree(w)
            sage: loads(dumps(s))
            Implicit Suffix Tree of the word: cacao
        """
        # For constructing the suffix tree.
        self._transition_function = {0: {}}
        self._suffix_link = {0: -1}
        self._active_state = (0, (1, 1))
        self._letters = []
        for letter in word:
            self._letters.append(letter)
            self._process_letter(letter)
        # _word is not needed for constructing the suffix tree,
        # but it is useful for the other methods.
        self._word = word

    def _process_letter(self, letter):
        r"""
        This is the main part of Ukkonen's algorithm.

        This corresponds to the algorithm "update" in [Ukko1995]_.

        .. NOTE::

           This function is a helper and does not update ``self._data`` and
           ``self._word``.

        REFERENCES:

        - [Ukko1995]_

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words("aco")("caca")
            sage: t = ImplicitSuffixTree(w); t
            Implicit Suffix Tree of the word: caca
            sage: new_letter = "o"
            sage: t._letters.append("o")
            sage: t._process_letter("o")
            sage: t._word = Words("aco")("cacao")
            sage: t
            Implicit Suffix Tree of the word: cacao

        ::

            sage: W = Words([0,1])
            sage: s = ImplicitSuffixTree(W([0,1,0,1])); s
            Implicit Suffix Tree of the word: 0101
            sage: s._letters.append(1)
            sage: s._process_letter(1)
            sage: s._word = W([0,1,0,1,1])
            sage: s
            Implicit Suffix Tree of the word: 01011
        """
        s, (k, i) = self._active_state
        old_r = 0
        end_state, r = self._test_and_split(s, (k, i-1), letter)
        while not end_state:
            # adjoin a new state rr and create a transition from r to rr
            rr = len(self._transition_function)
            self._transition_function[rr] = {}
            self._transition_function[r][(i, None)] = rr
            # update the suffix link, if necessary
            if old_r != 0:
                self._suffix_link[old_r] = r
            old_r = r
            # follow the suffix link to the next state
            s, k = self._canonize(self._suffix_link[s], (k, i-1))
            end_state, r = self._test_and_split(s, (k, i-1), letter)
        # update the suffix link, if necessary
        if old_r != 0:
            self._suffix_link[old_r] = s
        # set the active state
        s, k = self._canonize(s, (k, i))
        self._active_state = (s, (k, i+1))

    def _test_and_split(self, s, k_p, letter):
        r"""
        Helper function for :meth:`_process_letter`. Test to see
        whether an edge needs to be split.

        OUTPUT:

        ``(True, state)``, where ``state`` is the next state to
        process (either a newly created state or the original ``s``).

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words("aco")("caca")
            sage: t = ImplicitSuffixTree(w)
            sage: t._letters.append(w.parent().alphabet().rank("o"))
            sage: t._test_and_split(0, (4,5), w.parent().alphabet().rank("o"))
            (False, 3)
        """
        k, p = k_p
        if k <= p:
            # find the transition from s that begins with k-th letter
            (kk, pp), ss = self._find_transition(s, self._letters[k - 1])
            if letter == self._letters[kk + p - k]:
                return (True, s)
            else:
                # replace transition above by transitions
                del self._transition_function[s][(kk, pp)]
                r = len(self._transition_function)
                self._transition_function[r] = {}
                self._transition_function[s][(kk, kk+p-k)] = r
                self._transition_function[r][(kk+p-k+1, pp)] = ss
                return (False, r)
        else:
            transition = self._find_transition(s, letter)
            if transition is None:
                return (False, s)
            else:
                return (True, s)

    def _canonize(self, s, k_p):
        r"""
        Given an implicit or explicit reference pair for a node, return
        the canonical reference pair.

        Recall that a node r is referenced as (s, (k,p)), where s is an
        ancestor or r and w[k-1:p] is the word obtained by reading the edge
        labels along the path from s to r. A reference pair is canonical if
        s is the closest ancestor of r.

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: t = ImplicitSuffixTree(Word("cacao"))
            sage: t._canonize(0,(3,5))
            (3, 5)
            sage: t._canonize(0,(2,5))
            (5, 3)
        """
        k, p = k_p
        if p < k:
            return (s, k)
        else:
            (kk, pp), ss = self._find_transition(s, self._letters[k - 1])
            while pp is not None and pp - kk <= p - k:
                k = k + pp - kk + 1
                s = ss
                if k <= p:
                    (kk, pp), ss = self._find_transition(s, self._letters[k-1])
            return (s, k)

    def _find_transition(self, state, letter):
        r"""
        Return the transition from state that begins with letter.

        This returns ``None`` if no such transition exists.

        The transitions are stored as a dictionary of dictionaries: keyed
        by the nodes, with the corresponding dictionary keyed by pairs
        `(i,j)` of integers representing the word w[i-1:j].

        ._transition_function = {..., node: {(i,j): target_node, ...} }

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: t = ImplicitSuffixTree(Word("cacao"))
            sage: t._find_transition(-1, "c")
            ((0, 0), 0)
            sage: t._find_transition(0, "a")
            ((2, 2), 5)
            sage: t._find_transition(0, "c")
            ((1, 2), 3)
            sage: t._find_transition(5, "c")
            ((3, None), 2)
            sage: t._find_transition(5, "a")

        ::

            sage: t = ImplicitSuffixTree(Word([0,1,0,1,1]))
            sage: t._find_transition(3, 1)
            ((5, None), 4)
        """
        if state == -1:
            return ((0, 0), 0)

        if state in self._transition_function:
            for (k, p), s in self._transition_function[state].items():
                if self._letters[k - 1] == letter:
                    return ((k, p), s)
        return None

    # The following are not necessary for constructing the implicit suffix
    # tree; they add additional functionality to the class.

    # -------------
    # Visualization
    # -------------

    def _repr_(self) -> str:
        r"""
        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: ImplicitSuffixTree(Word("abcba"))._repr_()
            'Implicit Suffix Tree of the word: abcba'
        """
        return 'Implicit Suffix Tree of the %s' % repr(self.word())

    def word(self):
        r"""
        Return the word whose implicit suffix tree this is.

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: ImplicitSuffixTree(Word([0,1,0,1,0])).word() == Word([0,1,0,1,0])
            True
        """
        return self._word

    def transition_function_dictionary(self) -> dict:
        r"""
        Return the transition function as a dictionary of dictionaries.

        The format is consistent with the input format for ``DiGraph``.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: W = Words("aco")
            sage: t = ImplicitSuffixTree(W("cac"))
            sage: t.transition_function_dictionary()
            {0: {1: (0, None), 2: (1, None)}}

        ::

            sage: W = Words([0,1])
            sage: t = ImplicitSuffixTree(W([0,1,0]))
            sage: t.transition_function_dictionary()
            {0: {1: (0, None), 2: (1, None)}}
        """
        d = {}
        for u, v, (i, j) in self.edge_iterator():
            d.setdefault(u, {})[v] = (i, j)
        return d

    def to_digraph(self, word_labels=False):
        r"""
        Return a ``DiGraph`` object of the transition graph of the suffix tree.

        INPUT:

        - ``word_labels`` -- boolean (default: ``False``); if ``False``, labels
          the edges by pairs `(i, j)`. If ``True``, labels the edges by ``word[i:j]``

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: W = Words([0,1,2])
            sage: t = ImplicitSuffixTree(W([0,1,0,1,2]))
            sage: t.to_digraph()                                                        # needs sage.graphs
            Digraph on 8 vertices
        """
        if not self._letters:
            d = {0: {}}
            return DiGraph(d)
        d = self.transition_function_dictionary()
        for u in d:
            for v, (i, j) in d[u].items():
                if word_labels:
                    d[u][v] = self._word[i:j]
                elif j is None:
                    d[u][v] = (i, len(self._letters))
        return DiGraph(d)

    def plot(self, word_labels=False, layout='tree', tree_root=0,
             tree_orientation='up', vertex_colors=None, edge_labels=True,
             *args, **kwds):
        r"""
        Return a Graphics object corresponding to the transition graph of
        the suffix tree.

        INPUT:

        - ``word_labels`` -- boolean (default: ``False``); if ``False``, labels
          the edges by pairs `(i, j)`; if ``True``, labels the edges by ``word[i:j]``
        - ``layout`` -- (default: ``'tree'``)
        - ``tree_root`` -- (default: 0)
        - ``tree_orientation`` -- (default: ``'up'``)
        - ``vertex_colors`` -- (default: ``None``)
        - ``edge_labels`` -- (default: ``True``)

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: ImplicitSuffixTree(Word('cacao')).plot(word_labels=True)              # needs sage.graphs sage.plot
            Graphics object consisting of 23 graphics primitives
            sage: ImplicitSuffixTree(Word('cacao')).plot(word_labels=False)             # needs sage.graphs sage.plot
            Graphics object consisting of 23 graphics primitives

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: type(ImplicitSuffixTree(Word('cacao')).plot(word_labels=True))        # needs sage.graphs sage.plot
            <class 'sage.plot.graphics.Graphics'>
            sage: type(ImplicitSuffixTree(Word('cacao')).plot(word_labels=False))       # needs sage.graphs sage.plot
            <class 'sage.plot.graphics.Graphics'>
        """
        tree = self.to_digraph(word_labels=word_labels)
        if word_labels:
            for u, v, label in tree.edge_iterator():
                tree.set_edge_label(u, v, label.string_rep())
        if vertex_colors is None:
            vertex_colors = {'#fec7b8': tree.vertices(sort=True)}
        return tree.plot(layout=layout, tree_root=tree_root,
                         tree_orientation=tree_orientation,
                         vertex_colors=vertex_colors, edge_labels=edge_labels,
                         *args, **kwds)

    def show(self, word_labels=None, *args, **kwds):
        r"""
        Display the output of :meth:`plot`.

        INPUT:

        - ``word_labels`` -- (default: ``None``) if ``False``, labels the edges
          by pairs `(i, j)`; if ``True``, labels the edges by ``word[i:j]``

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words("cao")("cacao")
            sage: t = ImplicitSuffixTree(w)
            sage: t.show(word_labels=True)                                              # needs sage.plot
            sage: t.show(word_labels=False)                                             # needs sage.plot
        """
        self.plot(word_labels=word_labels, *args, **kwds).show()

    # ---------------
    # Various methods
    # ---------------

    def __eq__(self, other) -> bool:
        r"""
        If ``self`` and ``other`` have the same transition function
        and the same word, then they are equal.

        TESTS::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words([0,1,2])([0,1,0,1,2])
            sage: u = Words([0,1,2])(iter([0,1,0,1,2]))[:5]
            sage: ImplicitSuffixTree(w) == ImplicitSuffixTree(u)
            True
        """
        if not isinstance(other, ImplicitSuffixTree):
            return False
        return self._transition_function == other._transition_function \
            and self._letters == other._letters

    def transition_function(self, word, node=0):
        r"""
        Return the node obtained by starting from ``node`` and following the
        edges labelled by the letters of ``word``.

        OUTPUT:

        ``("explicit", end_node)`` if we end at ``end_node``, or
        ``("implicit", edge, d)`` if we end `d` spots along an edge.

        INPUT:

        - ``word`` -- a word
        - ``node`` -- (default: 0) starting node

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: W = Words([0,1,2])
            sage: t = ImplicitSuffixTree(W([0,1,0,1,2]))
            sage: t.transition_function(W([0,1,0]))
            ('implicit', (3, 1), 1)
            sage: t.transition_function(W([0,1,2]))
            ('explicit', 4)
            sage: t.transition_function(W([0,1,2]), 5)
            ('explicit', 2)
            sage: t.transition_function(W([0,1]), 5)
            ('implicit', (5, 2), 2)
        """
        if word.is_empty():
            return "explicit", node
        (k, p), s = self._find_transition(node, word[0])
        if p is None:
            # test that word is a prefix of self._letters[k-1:]
            if word == self._word[k-1:(k-1)+word.length()]:
                if word.length() == len(self._letters) - k + 1:
                    return "explicit", s
                else:
                    edge = (node, s)
                    return "implicit", edge, word.length()
        else:
            # find longest common prefix
            m = min(p-k+1, word.length())
            i = 0
            while i < m and self._word[k-1+i] == word[i]:
                i += 1
            if i == p-k+1:
                return self.transition_function(word[p-k+1:], s)
            else:
                edge = (node, s)
                return "implicit", edge, i
            return "explicit", node

    def states(self) -> list:
        r"""
        Return the states (explicit nodes) of the suffix tree.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: W = Words([0,1,2])
            sage: t = ImplicitSuffixTree(W([0,1,0,1,2]))
            sage: t.states()
            [0, 1, 2, 3, 4, 5, 6, 7]
        """
        return list(range(len(self._transition_function)))

    def suffix_link(self, state):
        r"""
        Evaluate the suffix link map of the implicit suffix tree on ``state``.

        Note that the suffix link is not defined for all states.

        The suffix link of a state `x'` that corresponds to the suffix `x` is
        defined to be -1 is `x'` is the root (0) and `y'` otherwise, where `y'`
        is the state corresponding to the suffix ``x[1:]``.

        INPUT:

        - ``state`` -- a state

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: W = Words([0,1,2])
            sage: t = ImplicitSuffixTree(W([0,1,0,1,2]))
            sage: t.suffix_link(3)
            5
            sage: t.suffix_link(5)
            0
            sage: t.suffix_link(0)
            -1
            sage: t.suffix_link(-1)
            Traceback (most recent call last):
                ...
            TypeError: there is no suffix link from -1
        """
        if state in self._suffix_link:
            return self._suffix_link[state]
        raise TypeError("there is no suffix link from %s" % state)

    def active_state(self):
        r"""
        Return the active state of the suffix tree.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: W = Words([0,1,2])
            sage: t = ImplicitSuffixTree(W([0,1,0,1,2]))
            sage: t.active_state()
            (0, (6, 6))
        """
        return self._active_state

    def process_letter(self, letter):
        r"""
        Modify the current implicit suffix tree producing the implicit
        suffix tree for ``self.word() + letter``.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words("aco")("cacao")
            sage: t = ImplicitSuffixTree(w[:-1]); t
            Implicit Suffix Tree of the word: caca
            sage: t.process_letter(w[-1]); t
            Implicit Suffix Tree of the word: cacao

        ::

            sage: W = Words([0,1])
            sage: s = ImplicitSuffixTree(W([0,1,0,1])); s
            Implicit Suffix Tree of the word: 0101
            sage: s.process_letter(W([1])[0]); s
            Implicit Suffix Tree of the word: 01011
        """
        self._word = self._word * self._word._parent([letter])
        self._letters.append(letter)
        self._process_letter(letter)

    def to_explicit_suffix_tree(self):
        r"""
        Convert ``self`` to an explicit suffix tree.

        It is obtained by processing an end of string letter as if it
        were a regular letter, except that no new leaf nodes are
        created (thus, the only thing that happens is that some
        implicit nodes become explicit).

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: w = Words("aco")("cacao")
            sage: t = ImplicitSuffixTree(w)
            sage: t.to_explicit_suffix_tree()

        ::

            sage: W = Words([0,1])
            sage: s = ImplicitSuffixTree(W([0,1,0,1,1]))
            sage: s.to_explicit_suffix_tree()
        """
        # append a new unique symbol to the word and process the new letter
        end_of_string = object()
        self._letters.append(end_of_string)
        s, (k, i) = self._active_state
        end_state, r = self._test_and_split(s, (k, i-1), end_of_string)
        while not end_state:
            s, k = self._canonize(self._suffix_link[s], (k, i-1))
            end_state, r = self._test_and_split(s, (k, i-1), end_of_string)
        # remove the end of string symbol from the word
        self._letters.pop()

    def edge_iterator(self):
        r"""
        Return an iterator over the edges of the suffix tree.

        The edge from `u` to `v` labelled by `(i,j)` is yielded as
        the tuple `(u,v,(i,j))`.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: sorted( ImplicitSuffixTree(Word("aaaaa")).edge_iterator() )
            [(0, 1, (0, None))]
            sage: sorted( ImplicitSuffixTree(Word([0,1,0,1])).edge_iterator() )
            [(0, 1, (0, None)), (0, 2, (1, None))]
            sage: sorted( ImplicitSuffixTree(Word()).edge_iterator() )
            []
        """
        queue = [0]
        while queue:
            v = queue.pop()
            for (i, j), u in self._transition_function[v].items():
                yield (v, u, (i - 1, j))
                queue.append(u)

    def number_of_factors(self, n=None):
        r"""
        Count the number of distinct factors of ``self.word()``.

        INPUT:

        - ``n`` -- integer or ``None``

        OUTPUT:

        If ``n`` is an integer, returns the number of distinct factors
        of length ``n``. If ``n`` is ``None``, returns the total number of
        distinct factors.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: t = ImplicitSuffixTree(Word([1,2,1,3,1,2,1]))
            sage: t.number_of_factors()
            22
            sage: t.number_of_factors(1)
            3
            sage: t.number_of_factors(9)
            0
            sage: t.number_of_factors(0)
            1

        ::

            sage: t = ImplicitSuffixTree(Word("cacao"))
            sage: t.number_of_factors()
            13
            sage: list(map(t.number_of_factors, range(10)))
            [1, 3, 3, 3, 2, 1, 0, 0, 0, 0]

        ::

            sage: t = ImplicitSuffixTree(Word("c"*1000))
            sage: t.number_of_factors()
            1001
            sage: t.number_of_factors(17)
            1
            sage: t.number_of_factors(0)
            1

        ::

            sage: ImplicitSuffixTree(Word()).number_of_factors()
            1

        ::

            sage: blueberry = ImplicitSuffixTree(Word("blueberry"))
            sage: blueberry.number_of_factors()
            43
            sage: list(map(blueberry.number_of_factors, range(10)))
            [1, 6, 8, 7, 6, 5, 4, 3, 2, 1]
        """
        if n is None:
            length_word = self.word().length()
            num_factors = 1  # empty word
            for u, v, (i, j) in self.edge_iterator():
                if j is None:
                    num_factors += length_word - i
                else:
                    num_factors += j - i
        elif isinstance(n, (int, Integer)):
            num_factors = 0
            queue = [(0, 0)]
            while queue:
                v, l = queue.pop()
                if l == n:
                    num_factors += 1
                if l < n:
                    if self._transition_function[v]:
                        for (i, j), u in self._transition_function[v].items():
                            if j is None:
                                j = self.word().length()
                            if j - i >= n - l:
                                num_factors += 1
                            else:
                                queue.append((u, l + j - i + 1))
        else:
            raise TypeError("not an integer or None: %s" % n)
        return num_factors

    def factor_iterator(self, n=None):
        r"""
        Generate distinct factors of ``self``.

        INPUT:

        - ``n`` -- integer or ``None``

        OUTPUT:

        If ``n`` is an integer, returns an iterator over all distinct
        factors of length ``n``. If ``n`` is ``None``, returns an iterator
        generating all distinct factors.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree
            sage: sorted( ImplicitSuffixTree(Word("cacao")).factor_iterator() )
            [word: , word: a, word: ac, word: aca, word: acao, word: ao, word: c, word: ca, word: cac, word: caca, word: cacao, word: cao, word: o]
            sage: sorted( ImplicitSuffixTree(Word("cacao")).factor_iterator(1) )
            [word: a, word: c, word: o]
            sage: sorted( ImplicitSuffixTree(Word("cacao")).factor_iterator(2) )
            [word: ac, word: ao, word: ca]
            sage: sorted( ImplicitSuffixTree(Word([0,0,0])).factor_iterator() )
            [word: , word: 0, word: 00, word: 000]
            sage: sorted( ImplicitSuffixTree(Word([0,0,0])).factor_iterator(2) )
            [word: 00]
            sage: sorted( ImplicitSuffixTree(Word([0,0,0])).factor_iterator(0) )
            [word: ]
            sage: sorted( ImplicitSuffixTree(Word()).factor_iterator() )
            [word: ]
            sage: sorted( ImplicitSuffixTree(Word()).factor_iterator(2) )
            []
        """
        # Every factor is a prefix of a suffix, so we do a depth
        # first search of the implicit suffix tree of the word.
        w = self.word()
        wlen = self.word().length()
        if n is None:
            queue = [(0, 0, -1, 0)]
            yield w[0:0]
            while queue:
                v, i, j, l = queue.pop()
                for k in range(i, j+1):
                    yield w[j-l:k]
                for (i, j), u in self._transition_function[v].items():
                    if j is None:
                        j = wlen
                    queue.append((u, i, j, l+j-i+1))
        elif isinstance(n, (int, Integer)):
            queue = [(0, 0, -1, 0)]
            while queue:
                v, i, j, l = queue.pop()
                if l == n:
                    yield w[j-l:j]
                if l < n:
                    for (i, j), u in self._transition_function[v].items():
                        if j is None:
                            j = wlen
                        if j - i >= n - l:
                            yield w[i-l-1:i-l+n-1]
                        else:
                            queue.append((u, i, j, l+j-i+1))
        else:
            raise TypeError("not an integer or None: %s" % n)

    def LZ_decomposition(self):
        r"""
        Return a list of index of the beginning of the block of the Lempel-Ziv
        decomposition of ``self.word``

        The *Lempel-Ziv decomposition* is the factorisation `u_1...u_k` of a
        word `w=x_1...x_n` such that `u_i` is the longest prefix of `u_i...u_k`
        that has an occurrence starting before `u_i` or a letter if this prefix
        is empty.

        OUTPUT:

        Return a list ``iB`` of index such that the blocks of the decomposition
        are ``self.word()[iB[k]:iB[k+1]]``

        EXAMPLES::

            sage: w = Word('abababb')
            sage: T = w.suffix_tree()
            sage: T.LZ_decomposition()
            [0, 1, 2, 6, 7]
            sage: w = Word('abaababacabba')
            sage: T = w.suffix_tree()
            sage: T.LZ_decomposition()
            [0, 1, 2, 3, 6, 8, 9, 11, 13]
            sage: w = Word([0, 0, 0, 1, 1, 0, 1])
            sage: T = w.suffix_tree()
            sage: T.LZ_decomposition()
            [0, 1, 3, 4, 5, 7]
            sage: w = Word('0000100101')
            sage: T = w.suffix_tree()
            sage: T.LZ_decomposition()
            [0, 1, 4, 5, 9, 10]
        """
        iB = [0]
        i = 0
        w = self.word()
        while i < len(w):
            l = 0
            (x, y), successor = self._find_transition(0, w[i])
            x = x-1
            while x < i+l:
                if y is None:
                    l = len(w)-i
                else:
                    l += y-x
                if i+l >= len(w):
                    l = len(w)-i
                    break
                (x, y), successor = self._find_transition(successor, w[i+l])
                x = x-1
            i += max(1, l)
            iB.append(i)
        return iB

    def _count_and_skip(self, node, i, j):
        r"""
        Use count and skip trick to follow the path starting at ``node`` and
        reading ``self.word()[i:j]``. We assume that reading
        ``self.word()[i:j]`` is possible from ``node``

        INPUT:

        - ``node`` -- explicit node of ``self``
        - ``i`` -- beginning of factor ``T.word()[i:j]``
        - ``j`` -- end of factor ``T.word()[i:j]``

        OUTPUT:

        The node obtained by starting at ``node`` and following the edges
        labeled by the letters of ``T.word()[i:j]``.
        Return ``("explicit", end_node)`` if w ends at the node ``end_node``,
        and ``("implicit", edge, d)`` if it ends after reading ``d`` letters along
        the edge ``edge``.

        EXAMPLES::

            sage: T = Word('00110111011').suffix_tree()
            sage: T._count_and_skip(5, 2, 5)
            ('implicit', (9, 10), 2)
            sage: T._count_and_skip(0, 1, 4)
            ('explicit', 7)
            sage: T._count_and_skip(0, 8, 10)
            ('implicit', (2, 7), 1)
            sage: T = Word('cacao').suffix_tree()
            sage: T._count_and_skip(3, 2, 5)
            ('explicit', 1)
        """
        trans = self._find_transition(node, self._letters[i])
        while trans[0][1] is not None and trans[0][1] - trans[0][0] + 1 <= j - i:
            node = trans[1]
            i += trans[0][1] - trans[0][0] + 1
            if i == j:
                return ('explicit', node)
            else:
                trans = self._find_transition(node, self._letters[i])
        if trans[0][1] is None and len(self.word()) - trans[0][0] + 1 <= j - i:
            return ('explicit', trans[1])
        else:
            return ('implicit', (node, trans[1]), j - i)

    def suffix_walk(self, edge, l):
        r"""
        Return the state of "w" if the input state is "aw".

        If the input state ``(edge, l)`` is path labeled "aw" with "a" a letter, the output is
        the state which is path labeled "w".

        INPUT:

        - ``edge`` -- the edge containing the state
        - ``l`` -- the string-depth of the state on edge (``l``>0)

        OUTPUT:

        Return ``("explicit", end_node)`` if the state of w is an explicit
        state and ``("implicit", edge, d)`` is obtained by reading ``d``
        letters on ``edge``.

        EXAMPLES::

            sage: T = Word('00110111011').suffix_tree()
            sage: T.suffix_walk((0, 5), 1)
            ('explicit', 0)
            sage: T.suffix_walk((7, 3), 1)
            ('implicit', (9, 4), 1)
        """
        start, end = edge
        # Select the transition that corresponds to edge
        ij = next(ij for ij, target in self._transition_function[start].items()
                  if target == end)

        # self.word()[i-1:j] is the word on the edges
        i = ij[0] - 1
        parent = self.suffix_link(start)
        return self._count_and_skip(parent, i, i + l)

    def leftmost_covering_set(self):
        r"""
        Compute the leftmost covering set of square pairs in ``self.word()``.
        Return a square as a pair ``(i,l)`` designating factor
        ``self.word()[i:i+l]``.

        A  leftmost covering set is a set such that the leftmost occurrence
        `(j,l)` of a square in ``self.word()`` is covered by a pair
        `(i,l)` in the set for all types of squares. We say that `(j,l)` is
        covered by `(i,l)` if `(i,l)` (i+1,l), \ldots, (j,l)` are all
        squares.

        The set is returned in the form of a list ``P`` such that ``P[i]``
        contains all the lengths of squares starting at ``i`` in the set.
        The lists ``P[i]`` are sorted in decreasing order.

        The algorithm used is described in [DS2004]_.

        EXAMPLES::

            sage: w = Word('abaabaabbaaabaaba')
            sage: T = w.suffix_tree()
            sage: T.leftmost_covering_set()
            [[6], [6], [2], [], [], [], [], [2], [], [], [6, 2], [], [], [], [], [], []]
            sage: w = Word('abaca')
            sage: T = w.suffix_tree()
            sage: T.leftmost_covering_set()
            [[], [], [], [], []]
            sage: T = Word('aaaaa').suffix_tree()
            sage: T.leftmost_covering_set()
            [[4, 2], [], [], [], []]
        """

        def condition1_square_pairs(i):
            r"""
            Compute the squares that have their center (the last letter of the
            first  occurrence of ``w`` in ``ww``) in the `i`-th block of the
            LZ-decomposition and that start in the `i`-th block and end in the
            `(i+1)`-th.
            """
            for k in range(1, B[i+1]-B[i]+1):
                q = B[i+1]-k
                k1 = w.longest_forward_extension(B[i+1], q) if B[i+1] < len(w) else 0
                k2 = w.longest_backward_extension(B[i+1]-1, q-1) if q > 0 else 0
                start = max(q-k2, q-k+1)
                if k1+k2 >= k and k1 > 0 and start >= B[i]:
                    yield (start, 2*k)

        def condition2_square_pairs(i):
            r"""
            Compute the squares that have their center (the last letter of the
            first  occurrence of ``w``  in ``ww``)  in the `i`-th  block of the
            LZ-decomposition and that starts in the `(i-1)`-th block or before.
            Their end is either in the `i`-th or the `(i+1)`-th block.
            """
            if i+2 < len(B):
                end = B[i+2] - B[i] + 1
            else:
                end = B[i+1] - B[i] + 1
            for k in range(2, end):
                q = B[i] + k
                k1 = w.longest_forward_extension(B[i], q) if q < len(w) else 0
                k2 = w.longest_backward_extension(B[i]-1, q-1) if B[i] > 0 else 0
                start = max(B[i]-k2, B[i]-k+1)
                if k1+k2 >= k and k1 > 0 and start+k <= B[i+1] and k2 > 0:
                    yield (start, 2*k)

        w = self.word()
        B = self.LZ_decomposition()
        P = [[] for _ in w]
        for i in range(len(B) - 1):
            for i, l in chain(condition2_square_pairs(i), condition1_square_pairs(i)):
                P[i].append(l)
        for l in P:
            l.reverse()
        return P

    # ---------------------
    # Miscellaneous methods
    # ---------------------

    def uncompactify(self):
        r"""
        Return the tree obtained from ``self`` by splitting edges so that they
        are labelled by exactly one letter.

        The resulting tree is isomorphic to the suffix trie.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree, SuffixTrie
            sage: abbab = Words("ab")("abbab")
            sage: s = SuffixTrie(abbab)
            sage: t = ImplicitSuffixTree(abbab)
            sage: t.uncompactify().is_isomorphic(s.to_digraph())                        # needs sage.graphs
            True
        """
        tree = self.to_digraph(word_labels=True)
        newtree = DiGraph()
        newtree.add_vertices(range(tree.order()))
        new_node = tree.order() + 1
        for u, v, label in tree.edge_iterator():
            if len(label) == 1:
                newtree.add_edge(u, v)
            else:
                newtree.add_edge(u, new_node, label[0])
                for w in label[1:-1]:
                    newtree.add_edge(new_node, new_node+1, w)
                    new_node += 1
                newtree.add_edge(new_node, v, label[-1])
                new_node += 1
        return newtree

    def trie_type_dict(self):
        r"""
        Return a dictionary in a format compatible with that of the suffix
        trie transition function.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import ImplicitSuffixTree, SuffixTrie
            sage: W = Words("ab")
            sage: t = ImplicitSuffixTree(W("aba"))
            sage: d = t.trie_type_dict()
            sage: len(d)
            5
            sage: d                     # random
            {(4, word: b): 5, (0, word: a): 4, (0, word: b): 3, (5, word: a): 1, (3, word: a): 2}
        """
        d = {}
        new_node = len(self._transition_function)
        for u, dd in self._transition_function.items():
            for sl, v in dd.items():
                w = self._word[sl[0]-1:sl[1]]
                if w.length() == 1:
                    d[u, w] = v
                else:
                    d[u, w[0:1]] = new_node
                    for i in range(1, w.length()-1):
                        d[new_node, w[i:i+1]] = new_node + 1
                        new_node += 1
                    d[new_node, w[-1:]] = v
                    new_node += 1
        return d


# ---------------------
# Decorated Suffix Tree
# ---------------------


class DecoratedSuffixTree(ImplicitSuffixTree):
    r"""
    The decorated suffix tree of a word.

    A *decorated suffix tree* of a word `w` is the suffix tree of `w`
    marked with the end point of all squares in the `w`.

    The symbol ``$`` is appended to ``w`` to ensure that each final
    state is a leaf of the suffix tree.

    INPUT:

    - ``w`` -- a finite word

    EXAMPLES::

        sage: from sage.combinat.words.suffix_trees import DecoratedSuffixTree
        sage: w = Word('0011001')
        sage: DecoratedSuffixTree(w)
        Decorated suffix tree of : 0011001$
        sage: w = Word('0011001', '01')
        sage: DecoratedSuffixTree(w)
        Decorated suffix tree of : 0011001$

    ALGORITHM:

    When using ``'pair'`` as output, the squares are retrieved in linear
    time. The algorithm is an implementation of the one proposed in
    [DS2004]_.
    """
    def __init__(self, w):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import DecoratedSuffixTree
            sage: w = Word('0011001')
            sage: DST = DecoratedSuffixTree(w)

        We skip the ``_test_and_split`` test because it is not a test meant
        for the ``TestSuite``::

            sage: TestSuite(DST).run(skip='_test_and_split')

        Test that we do not allow ``'$'`` to appear in the word::

            sage: w = Word('0011001$')
            sage: DecoratedSuffixTree(w)
            Traceback (most recent call last):
            ...
            ValueError: the symbol '$' is reserved for this class
        """
        if "$" in w:
            raise ValueError("the symbol '$' is reserved for this class")
        end_symbol = '$'
        w = Word(str(w) + end_symbol)
        ImplicitSuffixTree.__init__(self, w)
        self.labeling = self._complete_labeling()

    def __repr__(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import DecoratedSuffixTree
            sage: w = Word('0011001')
            sage: t = DecoratedSuffixTree(w)
            sage: t.__repr__()
            'Decorated suffix tree of : 0011001$'
        """
        w = self.word()
        if len(w) > 40:
            w = str(w[:40]) + '...'
        return f"Decorated suffix tree of : {w}"

    def _partial_labeling(self):
        r"""
        Make a depth-first search in the suffix tree and mark some squares of a
        leftmost covering set of the tree.

        This is used by :meth:`_complete_labeling`.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import DecoratedSuffixTree
            sage: w = Word('abaababbabba')
            sage: T = DecoratedSuffixTree(w)
            sage: T._partial_labeling()
            {(3, 4): [1], (5, 1): [3], (5, 6): [1], (11, 17): [1], (13, 8): [1], (15, 10): [2]}
        """
        def node_processing(node, parent, head):
            r"""
            Marks points along the edge ``(parent, node)`` if the string depth
            of parent is smaller than the length of the square at the head of
            ``P(node)``.
            Make it for all such square pairs and remove them from ``P(node)``.

            INPUT:

            - ``node`` -- a node of ``self``
            - ``parent`` -- the parent of a node in ``self``
            - ``head`` -- tuple indicating the head of the list ``P(node)``

            OUTPUT: ``(i, pos)``, the new head of ``P(node)``
            """
            i, pos = head
            pano = (parent, node)
            while pos < len(P[i]) and P[i][pos] > string_depth[parent]:
                label = P[i][pos] - string_depth[parent]
                if pano in labeling:
                    labeling[pano].append(label)
                else:
                    labeling[pano] = [label]
                pos += 1
            return (i, pos)

        def treat_node(current_node, parent):
            r"""
            Proceed to a depth-first search in ``self``, counting the
            string_depth of each node and processing each node for marking.

            To initiate the depth first search call ``self.treat_node(0,None)``

            INPUT:

            - ``current_node`` -- a node
            - ``parent`` -- parent of ``current_node`` in ``self``

            OUTPUT:

            The resulting list P(current_node) with current_node have been
            processed by ``node_processing``. The output is a pair ``(i,
            pos)`` such that ``P[i][pos:]`` is the list of current_node.
            """

            # Call recursively on children of current_node
            if current_node in D:
                node_list = (n, 0)
                for child in D[current_node]:
                    i, j = D[current_node][child]
                    if j is None:
                        j = n
                    string_depth[child] = string_depth[current_node] + j - i
                    child_list = treat_node(child, current_node)
                    if child_list[0] < node_list[0]:
                        node_list = child_list
            else:  # The node is a child
                node_list = (n - string_depth[current_node], 0)
            # Make treatment on current node head
            return node_processing(current_node, parent, node_list)

        P = self.leftmost_covering_set()
        D = self.transition_function_dictionary()
        string_depth = {0: 0}
        n = len(self.word())
        labeling = {}
        treat_node(0, None)
        return labeling

    def _complete_labeling(self):
        r"""
        Return a dictionary of edges of ``self``, with marked points for the end
        of each distinct squares that can be found in ``self.word()``.

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import DecoratedSuffixTree
            sage: w = Word('aabbaaba')
            sage: DecoratedSuffixTree(w)._complete_labeling()
            {(2, 7): [1], (5, 4): [1]}
        """

        def walk_chain(u, v, l, start):
            r"""
            Execute a chain of suffix walk until a walk is unsuccessful or it
            got to a point already registered in ``QP``. Registers all visited
            point in ``Q``.

            INPUT:

            - ``(u, v)`` -- edge on which the point is registered
            - ``l`` -- depth of the registered point on (u,v)
            - ``start`` -- beginning of the squares registered by the label
              ``(u, v), l``
            """
            # Mark the point in labeling
            if (u, v) in labeling:
                labeling[(u, v)].append(l)
            else:
                labeling[(u, v)] = [l]
            # Make the walk
            final_state = self.suffix_walk((u, v), l)
            successful = False
            if final_state[0] == 'explicit':
                parent = final_state[1]
                transition = self._find_transition(parent, self._letters[start])
                if transition is not None:
                    child = transition[1]
                    successful = True
                    depth = 1
            else:
                parent = final_state[1][0]
                child = final_state[1][1]
                depth = final_state[2]
                next_letter = self._letters[D[parent][child][0]+depth]
                if next_letter == self._letters[start]:
                    successful = True
                    depth += 1
            # If needed start a new walk
            if successful:
                if (parent, child) in prelabeling:
                    if depth not in prelabeling[(parent, child)]:
                        walk_chain(parent, child, depth, start+1)
                else:
                    walk_chain(parent, child, depth, start+1)

        def treat_node(current_node, i, j):
            r"""
            Execute a depth-first search on ``self`` and start a suffix walk
            for labeled points on each edges of T.

            The function is recursive, call
            ``treat_node(0,0,0)`` to initiate the search.

            INPUT:

            - ``current_node`` -- the node to treat
            - ``(i, j)`` -- pair of index such that the path from 0 to
              ``current_node`` reads ``self.word()[i:j]``
            """

            if current_node in D:
                for child in D[current_node]:
                    edge = (current_node, child)
                    edge_label = D[edge[0]][edge[1]]
                    treat_node(child, edge_label[0]-(j-i), edge_label[1])
                    if (current_node, child) in prelabeling:
                        for l in prelabeling[edge]:
                            square_start = edge_label[0] - (j - i)
                            walk_chain(current_node, child, l, square_start)

        prelabeling = self._partial_labeling()
        labeling = {}
        D = self.transition_function_dictionary()
        treat_node(0, 0, 0)
        return labeling

    def square_vocabulary(self, output='pair'):
        r"""
        Return the list of distinct squares of ``self.word``.

        Two types of outputs are available `pair` and `word`. The algorithm
        is only truly linear if `output` is set to `pair`. A pair is a tuple
        `(i, l)` that indicates the factor ``self.word()[i:i+l]``.
        The option ``'word'`` return word objects.

        INPUT:

        - ``output`` -- (default: ``'pair'``) either ``'pair'`` or ``'word'``

        EXAMPLES::

            sage: from sage.combinat.words.suffix_trees import DecoratedSuffixTree
            sage: w = Word('aabb')
            sage: sorted(DecoratedSuffixTree(w).square_vocabulary())
            [(0, 0), (0, 2), (2, 2)]
            sage: w = Word('00110011010')
            sage: sorted(DecoratedSuffixTree(w).square_vocabulary(output='word'))
            [word: , word: 00, word: 00110011, word: 01100110, word: 1010, word: 11]
        """
        def treat_node(current_node, i, j):
            if current_node in D:
                for child in D[current_node]:
                    edge = (current_node, child)
                    edge_label = (D[edge[0]][edge[1]])
                    treat_node(child, edge_label[0]-(j-i), edge_label[1])
                    if (current_node, child) in Q:
                        for l in Q[(current_node, child)]:
                            square_start = edge_label[0]-(j-i)
                            pair = (square_start, edge_label[0]+l-square_start)
                            squares.append(pair)

        if output not in ["pair", "word"]:
            msg = f"output should be 'pair' or 'word'; got {output}"
            raise ValueError(msg)
        D = self.transition_function_dictionary()
        Q = self.labeling
        squares = [(0, 0)]
        treat_node(0, 0, 0)
        if output == "pair":
            return squares
        return [self.word()[i:i + l] for i, l in squares]



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
r"""
Jack symmetric functions

Jack's symmetric functions appear in [Ma1995]_ Chapter VI, section 10.
Zonal polynomials are the subject of [Ma1995]_ Chapter VII.
The parameter `\alpha` in that reference is the parameter `t` in this
implementation in sage.

REFERENCES:

.. [Jack1970] \H. Jack,
   *A class of symmetric functions with a parameter*,
   Proc. R. Soc. Edinburgh (A), 69, 1-18.

.. [Ma1995] \I. G. Macdonald,
   *Symmetric functions and Hall polynomials*,
   second ed.,
   The Clarendon Press, Oxford University Press, New York, 1995, With contributions
   by A. Zelevinsky, Oxford Science Publications.
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>
#                     2012 Mike Zabrocki <mike.zabrocki@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.categories.all
from sage.arith.functions import lcm
from sage.arith.misc import gcd
from sage.categories.homset import End, Hom
from sage.categories.morphism import SetMorphism
from sage.misc.misc_c import prod
from sage.rings.fraction_field import FractionField, FractionField_generic
from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.structure.unique_representation import UniqueRepresentation

from . import sfa

QQt = FractionField(QQ['t'])

p_to_m_cache = {}
m_to_p_cache = {}


class Jack(UniqueRepresentation):
    @staticmethod
    def __classcall__(cls, Sym, t='t'):
        """
        Normalize the arguments.

        TESTS::

            sage: R.<q, t> = QQ[]
            sage: B1 = SymmetricFunctions(R).jack().P()
            sage: B2 = SymmetricFunctions(R).jack(t).P()
            sage: B3 = SymmetricFunctions(R).jack(q).P()
            sage: B1 is B2
            True
            sage: B1 == B3
            False
        """
        return super().__classcall__(cls, Sym, Sym.base_ring()(t))

    def __init__(self, Sym, t):
        r"""
        The family of Jack symmetric functions including the `P`, `Q`, `J`, `Qp`
        bases.  The default parameter is ``t``.

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases
        - ``Sym`` -- a ring of symmetric functions
        - ``t`` -- an optional parameter (default: ``'t'``)

        EXAMPLES::

            sage: SymmetricFunctions(FractionField(QQ['t'])).jack()
            Jack polynomials over Fraction Field of Univariate Polynomial Ring in t over Rational Field
            sage: SymmetricFunctions(QQ).jack(1)
            Jack polynomials with t=1 over Rational Field
        """
        self._sym = Sym
        self.t = t
        self._name_suffix = ""
        if str(t) != 't':
            self._name_suffix += " with t=%s" % t
        self._name = "Jack polynomials"+self._name_suffix+" over "+repr(Sym.base_ring())

    def __repr__(self):
        r"""
        The string representation for the family of Jack symmetric function bases.

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases

        OUTPUT: the name of the family of bases

        EXAMPLES::

            sage: SymmetricFunctions(QQ).jack(1)
            Jack polynomials with t=1 over Rational Field
        """
        return self._name

    def base_ring( self ):
        r"""
        Return the base ring of the symmetric functions in which the
        Jack symmetric functions live.

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases

        OUTPUT: the base ring of the symmetric functions ring of ``self``

        EXAMPLES::

            sage: J2 = SymmetricFunctions(QQ).jack(t=2)
            sage: J2.base_ring()
            Rational Field
        """
        return self._sym.base_ring()

    def symmetric_function_ring( self ):
        r"""
        Return the base ring of the symmetric functions of the Jack symmetric
        function bases

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases

        OUTPUT: the symmetric functions ring of ``self``

        EXAMPLES::

            sage: Jacks = SymmetricFunctions(FractionField(QQ['t'])).jack()
            sage: Jacks.symmetric_function_ring()
            Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field
        """
        return self._sym

    def P(self):
        r"""
        Return the algebra of Jack polynomials in the `P` basis.

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases

        OUTPUT: the `P` basis of the Jack symmetric functions

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JP = Sym.jack().P(); JP
            Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the Jack P basis
            sage: Sym.jack(t=-1).P()
            Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the Jack P with t=-1 basis

        At `t = 1`, the Jack polynomials in the `P` basis are the Schur
        symmetric functions.

        ::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JP = Sym.jack(t=1).P()
            sage: s = Sym.schur()
            sage: s(JP([2,2,1]))
            s[2, 2, 1]
            sage: JP(s([2,2,1]))
            JackP[2, 2, 1]
            sage: JP([2,1])^2
            JackP[2, 2, 1, 1] + JackP[2, 2, 2] + JackP[3, 1, 1, 1] + 2*JackP[3, 2, 1] + JackP[3, 3] + JackP[4, 1, 1] + JackP[4, 2]

        At `t = 2`, the Jack polynomials in the `P` basis are the zonal
        polynomials.

        ::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JP = Sym.jack(t=2).P()
            sage: Z = Sym.zonal()
            sage: Z(JP([2,2,1]))
            Z[2, 2, 1]
            sage: JP(Z[2, 2, 1])
            JackP[2, 2, 1]
            sage: JP([2])^2
            64/45*JackP[2, 2] + 16/21*JackP[3, 1] + JackP[4]
            sage: Z([2])^2
            64/45*Z[2, 2] + 16/21*Z[3, 1] + Z[4]

        ::

            sage: Sym = SymmetricFunctions(QQ['a','b'].fraction_field())
            sage: (a,b) = Sym.base_ring().gens()
            sage: Jacka = Sym.jack(t=a)
            sage: Jackb = Sym.jack(t=b)
            sage: m = Sym.monomial()
            sage: JPa = Jacka.P()
            sage: JPb = Jackb.P()
            sage: m(JPa[2,1])
            (6/(a+2))*m[1, 1, 1] + m[2, 1]
            sage: m(JPb[2,1])
            (6/(b+2))*m[1, 1, 1] + m[2, 1]
            sage: m(a*JPb([2,1]) + b*JPa([2,1]))
            ((6*a^2+6*b^2+12*a+12*b)/(a*b+2*a+2*b+4))*m[1, 1, 1] + (a+b)*m[2, 1]
            sage: JPa(JPb([2,1]))
            ((6*a-6*b)/(a*b+2*a+2*b+4))*JackP[1, 1, 1] + JackP[2, 1]

        ::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JQ = Sym.jack().Q()
            sage: JP = Sym.jack().P()
            sage: JJ = Sym.jack().J()

        ::

            sage: JP(JQ([2,1]))
            ((1/2*t+1)/(t^3+1/2*t^2))*JackP[2, 1]
            sage: JP(JQ([3]))
            ((1/3*t^2+1/2*t+1/6)/t^3)*JackP[3]
            sage: JP(JQ([1,1,1]))
            (6/(t^3+3*t^2+2*t))*JackP[1, 1, 1]

        ::

            sage: JP(JJ([3]))
            (2*t^2+3*t+1)*JackP[3]
            sage: JP(JJ([2,1]))
            (t+2)*JackP[2, 1]
            sage: JP(JJ([1,1,1]))
            6*JackP[1, 1, 1]

        ::

            sage: s = Sym.schur()
            sage: JP(s([2,1]))
            ((2*t-2)/(t+2))*JackP[1, 1, 1] + JackP[2, 1]
            sage: s(_)
            s[2, 1]
        """
        return JackPolynomials_p(self)

    def Q(self):
        r"""
        Return the algebra of Jack polynomials in the `Q` basis.

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases

        OUTPUT: the `Q` basis of the Jack symmetric functions

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JQ = Sym.jack().Q(); JQ
            Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the Jack Q basis
            sage: Sym = SymmetricFunctions(QQ)
            sage: Sym.jack(t=-1).Q()
            Symmetric Functions over Rational Field in the Jack Q with t=-1 basis

        ::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JQ = Sym.jack().Q()
            sage: JP = Sym.jack().P()
            sage: JQ(sum(JP(p) for p in Partitions(3)))
            (1/6*t^3+1/2*t^2+1/3*t)*JackQ[1, 1, 1] + ((2*t^3+t^2)/(t+2))*JackQ[2, 1] + (3*t^3/(t^2+3/2*t+1/2))*JackQ[3]

        ::

            sage: s = Sym.schur()
            sage: JQ(s([3])) # indirect doctest
            (1/6*t^3-1/2*t^2+1/3*t)*JackQ[1, 1, 1] + ((2*t^3-2*t^2)/(t+2))*JackQ[2, 1] + (3*t^3/(t^2+3/2*t+1/2))*JackQ[3]
            sage: JQ(s([2,1]))
            (1/3*t^3-1/3*t)*JackQ[1, 1, 1] + ((2*t^3+t^2)/(t+2))*JackQ[2, 1]
            sage: JQ(s([1,1,1]))
            (1/6*t^3+1/2*t^2+1/3*t)*JackQ[1, 1, 1]
        """
        return JackPolynomials_q(self)

    def J(self):
        r"""
        Return the algebra of Jack polynomials in the `J` basis.

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases

        OUTPUT: the `J` basis of the Jack symmetric functions

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JJ = Sym.jack().J(); JJ
            Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the Jack J basis
            sage: Sym = SymmetricFunctions(QQ)
            sage: Sym.jack(t=-1).J()
            Symmetric Functions over Rational Field in the Jack J with t=-1 basis

        At `t = 1`, the Jack polynomials in the `J` basis are scalar multiples
        of the Schur functions with the scalar given by a Partition's
        :meth:`~sage.combinat.partition.Partition.hook_product` method at 1::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JJ = Sym.jack(t=1).J()
            sage: s = Sym.schur()
            sage: p = Partition([3,2,1,1])
            sage: s(JJ(p)) == p.hook_product(1)*s(p)  # long time (4s on sage.math, 2012)
            True

        At `t = 2`, the Jack polynomials in the `J` basis are scalar multiples
        of the zonal polynomials with the scalar given by a Partition's
        :meth:`~sage.combinat.partition.Partition.hook_product` method at 2.

        ::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JJ = Sym.jack(t=2).J()
            sage: Z = Sym.zonal()
            sage: p = Partition([2,2,1])
            sage: Z(JJ(p)) == p.hook_product(2)*Z(p)
            True

        ::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JJ = Sym.jack().J()
            sage: JP = Sym.jack().P()
            sage: JJ(sum(JP(p) for p in Partitions(3)))
            1/6*JackJ[1, 1, 1] + (1/(t+2))*JackJ[2, 1] + (1/2/(t^2+3/2*t+1/2))*JackJ[3]

        ::

            sage: s = Sym.schur()
            sage: JJ(s([3])) # indirect doctest
            ((1/6*t^2-1/2*t+1/3)/(t^2+3*t+2))*JackJ[1, 1, 1] + ((t-1)/(t^2+5/2*t+1))*JackJ[2, 1] + (1/2/(t^2+3/2*t+1/2))*JackJ[3]
            sage: JJ(s([2,1]))
            ((1/3*t-1/3)/(t+2))*JackJ[1, 1, 1] + (1/(t+2))*JackJ[2, 1]
            sage: JJ(s([1,1,1]))
            1/6*JackJ[1, 1, 1]
        """
        return JackPolynomials_j(self)

    def Qp(self):
        r"""
        Return the algebra of Jack polynomials in the `Qp`, which is dual to
        the `P` basis with respect to the standard scalar product.

        INPUT:

        - ``self`` -- the family of Jack symmetric function bases

        OUTPUT: the `Q'` basis of the Jack symmetric functions

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JP = Sym.jack().P()
            sage: JQp = Sym.jack().Qp(); JQp
            Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the Jack Qp basis
            sage: a = JQp([2])
            sage: a.scalar(JP([2]))
            1
            sage: a.scalar(JP([1,1]))
            0
            sage: JP(JQp([2]))                        # todo: missing auto normalization
            ((t-1)/(t+1))*JackP[1, 1] + JackP[2]
            sage: JP._normalize(JP(JQp([2])))
            ((t-1)/(t+1))*JackP[1, 1] + JackP[2]
        """
        return JackPolynomials_qp(self)

###################################################################


def c1(part, t):
    r"""
    Return the `t`-Jack scalar product between ``J(part)`` and ``P(part)``.

    INPUT:

    - ``part`` -- a partition
    - ``t`` -- an optional parameter (default: uses the parameter `t` from the
      Jack basis)

    OUTPUT:

    - a polynomial in the parameter ``t`` which is equal to the scalar
      product of ``J(part)`` and ``P(part)``

    EXAMPLES::

        sage: from sage.combinat.sf.jack import c1
        sage: t = QQ['t'].gen()
        sage: [c1(p,t) for p in Partitions(3)]
        [2*t^2 + 3*t + 1, t + 2, 6]
    """
    return prod([1+t*part.arm_lengths(flat=True)[i]+part.leg_lengths(flat=True)[i] for i in range(sum(part))],
                t.parent().one())


def c2(part, t):
    r"""
    Return the t-Jack scalar product between ``J(part)`` and ``Q(part)``.

    INPUT:

    - ``self`` -- a Jack basis of the symmetric functions
    - ``part`` -- a partition
    - ``t`` -- an optional parameter (default: uses the parameter `t` from the
      Jack basis)

    OUTPUT:

    - a polynomial in the parameter ``t`` which is equal to the scalar
      product of ``J(part)`` and ``Q(part)``

    EXAMPLES::

        sage: from sage.combinat.sf.jack import c2
        sage: t = QQ['t'].gen()
        sage: [c2(p,t) for p in Partitions(3)]
        [6*t^3, 2*t^3 + t^2, t^3 + 3*t^2 + 2*t]
    """
    return prod([t+t*part.arm_lengths(flat=True)[i]+part.leg_lengths(flat=True)[i] for i in range(sum(part))],
                t.parent().one())


def normalize_coefficients(self, c):
    r"""
    If our coefficient ring is the field of fractions over a univariate
    polynomial ring over the rationals, then we should clear both the
    numerator and denominator of the denominators of their
    coefficients.

    INPUT:

    - ``self`` -- a Jack basis of the symmetric functions
    - ``c`` -- a coefficient in the base ring of ``self``

    OUTPUT: divide numerator and denominator by the greatest common divisor

    EXAMPLES::

        sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
        sage: t = JP.base_ring().gen()
        sage: a = 2/(1/2*t+1/2)
        sage: JP._normalize_coefficients(a)
        4/(t + 1)
        sage: a = 1/(1/3+1/6*t)
        sage: JP._normalize_coefficients(a)
        6/(t + 2)
        sage: a = 24/(4*t^2 + 12*t + 8)
        sage: JP._normalize_coefficients(a)
        6/(t^2 + 3*t + 2)
    """
    BR = self.base_ring()
    if isinstance(BR, FractionField_generic) and BR.base_ring() == QQ:
        denom = c.denominator()
        numer = c.numerator()

        #Clear the denominators
        a = lcm([i.denominator() for i in denom.coefficients(sparse=False)])
        b = lcm([i.denominator() for i in numer.coefficients(sparse=False)])
        l = Integer(a).lcm(Integer(b))
        denom *= l
        numer *= l

        #Divide through by the gcd of the numerators
        a = gcd([i.numerator() for i in denom.coefficients(sparse=False)])
        b = gcd([i.numerator() for i in numer.coefficients(sparse=False)])
        l = Integer(a).gcd(Integer(b))

        denom = denom // l
        numer = numer // l

        return c.parent()(numer, denom)
    else:
        return c

####################################################################


class JackPolynomials_generic(sfa.SymmetricFunctionAlgebra_generic):
    def __init__(self, jack):
        r"""
        A class of methods which are common to all Jack bases of the symmetric functions.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``jack`` -- a family of Jack symmetric function bases

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JP = Sym.jack().P(); JP.base_ring()
            Fraction Field of Univariate Polynomial Ring in t over Rational Field
            sage: Sym = SymmetricFunctions(QQ)
            sage: JP = Sym.jack(t=2).P(); JP.base_ring()
            Rational Field
        """
        s = self.__class__.__name__[16:].capitalize()
        sfa.SymmetricFunctionAlgebra_generic.__init__(
            self, jack._sym,
            basis_name="Jack " + s + jack._name_suffix,
            prefix="Jack" + s)
        self.t = jack.t
        self._sym = jack._sym
        self._jack = jack

        # Bases defined by orthotriangularity should inherit from some
        # common category BasesByOrthotriangularity (shared with Jack, HL, orthotriang, Mcdo)
        if hasattr(self, "_m_cache"):
            # temporary until Hom(GradedHopfAlgebrasWithBasis work better)
            category = sage.categories.all.ModulesWithBasis(self._sym.base_ring())
            self._m = self._sym.monomial()
            self   .register_coercion(SetMorphism(Hom(self._m, self, category), self._m_to_self))
            self._m.register_coercion(SetMorphism(Hom(self, self._m, category), self._self_to_m))
        if hasattr(self, "_h_cache"):
            # temporary until Hom(GradedHopfAlgebrasWithBasis work better)
            category = sage.categories.all.ModulesWithBasis(self._sym.base_ring())
            self._h = self._sym.homogeneous()
            self   .register_coercion(SetMorphism(Hom(self._h, self, category), self._h_to_self))
            self._h.register_coercion(SetMorphism(Hom(self, self._h, category), self._self_to_h))

    def construction(self):
        """
        Return a pair ``(F, R)``, where ``F`` is a
        :class:`SymmetricFunctionsFunctor` and `R` is a ring, such
        that ``F(R)`` returns ``self``.

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: JP = Sym.jack().P()
            sage: JP.construction()
            (SymmetricFunctionsFunctor[Jack P],
             Fraction Field of Univariate Polynomial Ring in t over Rational Field)
        """
        return (sfa.SymmetricFunctionsFamilyFunctor(self, Jack,
                                                    self.basis_name(),
                                                    self.t),
                self.base_ring())

    def _m_to_self(self, x):
        r"""
        Isomorphism from the monomial basis into ``self``.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``x`` -- element of the monomial basis

        OUTPUT: an element of ``self`` equivalent to ``x``

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JP = Sym.jack(t=2).P()
            sage: m = Sym.monomial()
            sage: JP._m_to_self(m[2,1])
            -3/2*JackP[1, 1, 1] + JackP[2, 1]

        This is for internal use only. Please use instead::

            sage: JP(m[2,1])
            -3/2*JackP[1, 1, 1] + JackP[2, 1]
        """
        return self._from_cache(x, self._m_cache, self._m_to_self_cache,
                                t=self.t)

    def _self_to_m(self, x):
        r"""
        Isomorphism from ``self`` to the monomial basis.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``x`` -- an element of ``self``

        OUTPUT: an element of the monomial basis equivalent to ``x``

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JP = Sym.jack(t=2).P()
            sage: m = Sym.monomial()
            sage: JP._self_to_m(JP[2,1])
            3/2*m[1, 1, 1] + m[2, 1]

        This is for internal use only. Please use instead::

            sage: m(JP[2,1])
            3/2*m[1, 1, 1] + m[2, 1]
        """
        return self._m._from_cache(x, self._m_cache, self._self_to_m_cache,
                                   t=self.t)

    def c1(self, part):
        r"""
        Return the `t`-Jack scalar product between ``J(part)`` and ``P(part)``.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``part`` -- a partition
        - ``t`` -- an optional parameter (default: uses the parameter `t` from the
          Jack basis)

        OUTPUT:

        - a polynomial in the parameter ``t`` which is equal to the scalar
          product of ``J(part)`` and ``P(part)``

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: JP.c1(Partition([2,1]))
            t + 2
        """
        return c1(part, self.t)

    def c2(self, part):
        r"""
        Return the `t`-Jack scalar product between ``J(part)`` and ``Q(part)``.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``part`` -- a partition
        - ``t`` -- an optional parameter (default: uses the parameter `t` from the
            Jack basis)

        OUTPUT:

        - a polynomial in the parameter ``t`` which is equal to the scalar
          product of ``J(part)`` and ``Q(part)``

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: JP.c2(Partition([2,1]))
            2*t^3 + t^2
        """
        return c2(part, self.t)

    _normalize_coefficients = normalize_coefficients

    def _normalize(self, x):
        r"""
        Normalize the coefficients of ``x``.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``x`` -- an element of ``self``

        OUTPUT: ``x`` with _normalize_coefficient applied to each of the coefficients

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: t = JP.base_ring().gen()
            sage: a = 2/(1/2*t+1/2)
            sage: b = 1/(1/3+1/6*t)
            sage: c = 24/(4*t^2 + 12*t + 8)
            sage: JP._normalize( a*JP[1] + b*JP[2] + c*JP[2,1] )
            (4/(t+1))*JackP[1] + (6/(t+2))*JackP[2] + (6/(t^2+3*t+2))*JackP[2, 1]

        .. TODO:: this should be a method on the elements (what's the standard name for such methods?)
        """
        return x.map_coefficients(self._normalize_coefficients)

    def _normalize_morphism(self, category):
        r"""
        Return the normalize morphism.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``category`` -- a category

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: normal = JP._normalize_morphism(AlgebrasWithBasis(JP.base_ring()))
            sage: normal.parent()
            Set of Homomorphisms from Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the Jack P basis to Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the Jack P basis
            sage: normal.category_for()
            Category of algebras with basis over Fraction Field of Univariate Polynomial Ring in t over Rational Field

            sage: t = JP.t
            sage: a = 2/(1/2*t+1/2)
            sage: b = 1/(1/3+1/6*t)
            sage: c = 24/(4*t^2 + 12*t + 8)
            sage: normal( a*JP[1] + b*JP[2] + c*JP[2,1] )
            (4/(t+1))*JackP[1] + (6/(t+2))*JackP[2] + (6/(t^2+3*t+2))*JackP[2, 1]

        .. TODO::

            This method should not be needed once short idioms to
            construct morphisms are available
        """
        return SetMorphism(End(self, category), self._normalize)

    def product(self, left, right):
        r"""
        The product of two Jack symmetric functions is done by multiplying the
        elements in the `P` basis and then expressing the elements
        in the basis ``self``.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``left``, ``right`` -- symmetric function elements

        OUTPUT: the product of ``left`` and ``right`` expanded in the basis ``self``

        EXAMPLES::

            sage: JJ = SymmetricFunctions(FractionField(QQ['t'])).jack().J()
            sage: JJ([1])^2              # indirect doctest
            (t/(t+1))*JackJ[1, 1] + (1/(t+1))*JackJ[2]
            sage: JJ([2])^2
            (t^2/(t^2+3/2*t+1/2))*JackJ[2, 2] + (4/3*t/(t^2+4/3*t+1/3))*JackJ[3, 1] + ((1/6*t+1/6)/(t^2+5/6*t+1/6))*JackJ[4]
            sage: JQ = SymmetricFunctions(FractionField(QQ['t'])).jack().Q()
            sage: JQ([1])^2              # indirect doctest
            JackQ[1, 1] + (2/(t+1))*JackQ[2]
            sage: JQ([2])^2
            JackQ[2, 2] + (2/(t+1))*JackQ[3, 1] + ((t+1)/(t^2+5/6*t+1/6))*JackQ[4]
        """
        return self(self._P(left) * self._P(right))

    def jack_family(self):
        r"""
        Return the family of Jack bases associated to the basis ``self``.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions

        EXAMPLES::

            sage: JackP = SymmetricFunctions(QQ).jack(t=2).P()
            sage: JackP.jack_family()
            Jack polynomials with t=2 over Rational Field
        """
        return self._jack

    def coproduct_by_coercion(self, elt):
        r"""
        Return the coproduct of the element ``elt`` by coercion to the Schur basis.

        INPUT:

        - ``self`` -- a Jack symmetric function basis
        - ``elt`` -- an instance of this basis

        OUTPUT:

        - The coproduct acting on ``elt``, the result is an element of the
          tensor squared of the Jack symmetric function basis

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ['t'].fraction_field())
            sage: Sym.jack().P()[2,2].coproduct()  # indirect doctest                   # needs lrcalc_python
            JackP[] # JackP[2, 2] + (2/(t+1))*JackP[1] # JackP[2, 1] + ((8*t+4)/(t^3+4*t^2+5*t+2))*JackP[1, 1] # JackP[1, 1] + JackP[2] # JackP[2] + (2/(t+1))*JackP[2, 1] # JackP[1] + JackP[2, 2] # JackP[]
        """
        from sage.categories.tensor import tensor
        s = self.realization_of().schur()
        g = self.tensor_square().sum(coeff*tensor([self(s[x]), self(s[y])])
                                        for ((x,y), coeff) in s(elt).coproduct())
        normalize = self._normalize_coefficients
        return self.tensor_square().sum(normalize(coeff)*tensor([self(x), self(y)])
                    for ((x,y), coeff) in g)

    class Element(sfa.SymmetricFunctionAlgebra_generic.Element):
        def scalar_jack(self, x, t=None):
            r"""
            A scalar product where the power sums are orthogonal and
            `\langle p_\mu, p_\mu \rangle = z_\mu t^{length(\mu)}`

            INPUT:

            - ``self`` -- an element of a Jack basis of the symmetric functions
            - ``x`` -- an element of the symmetric functions
            - ``t`` -- an optional parameter (default: ``None``; uses the
              parameter from the basis)

            OUTPUT: the Jack scalar product between ``x`` and ``self``

            EXAMPLES::

                sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
                sage: JP = Sym.jack().P()
                sage: JQ = Sym.jack().Q()
                sage: p = Partitions(3).list()
                sage: matrix([[JP(a).scalar_jack(JQ(b)) for a in p] for b in p])
                [1 0 0]
                [0 1 0]
                [0 0 1]
            """
            parent = self.parent()
            p = parent.realization_of().power()
            res = p(self).scalar_jack(p(x), t)

            return parent._normalize_coefficients(res)


def part_scalar_jack(part1, part2, t):
    r"""
    Return the Jack scalar product between ``p(part1)`` and ``p(part2)`` where
    `p` is the power-sum basis.

    INPUT:

    - ``part1``, ``part2`` -- two partitions
    - ``t`` -- a parameter

    OUTPUT: the scalar product between the power sum indexed by ``part1`` and ``part2``

    EXAMPLES::

        sage: Q.<t> = QQ[]
        sage: from sage.combinat.sf.jack import part_scalar_jack
        sage: matrix([[part_scalar_jack(p1,p2,t) for p1 in Partitions(4)] for p2 in Partitions(4)])
        [   4*t      0      0      0      0]
        [     0  3*t^2      0      0      0]
        [     0      0  8*t^2      0      0]
        [     0      0      0  4*t^3      0]
        [     0      0      0      0 24*t^4]
    """
    if part1 != part2:
        return 0
    else:
        return part1.centralizer_size()*t**len(part1)

#P basis


class JackPolynomials_p(JackPolynomials_generic):

    def __init__(self, jack):
        r"""
        The `P` basis is uni-triangularly related to the monomial basis and
        orthogonal with respect to the Jack scalar product.

        INPUT:

        - ``self`` -- an instance of the Jack `P` basis of the symmetric functions
        - ``jack`` -- a family of Jack symmetric function bases

        EXAMPLES::

            sage: # needs lrcalc_python
            sage: P = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: TestSuite(P).run(skip=['_test_associativity', '_test_distributivity', '_test_prod']) # products are too expensive
            sage: TestSuite(P).run(elements = [P.t*P[1,1]+P[2], P[1]+(1+P.t)*P[1,1]])
        """
        self._name = "Jack polynomials in the P basis"
        self._prefix = "JackP"

        self._m_to_self_cache = m_to_p_cache
        self._self_to_m_cache = p_to_m_cache
        JackPolynomials_generic.__init__(self, jack)

    def _m_cache(self, n):
        r"""
        Compute the change of basis between the Jack polynomials in the `P`
        basis and the monomial symmetric functions. This uses Gram-Schmidt
        to go to the monomials, and then that matrix is simply inverted.

        INPUT:

        - ``self`` -- an instance of the Jack `P` basis of the symmetric functions
        - ``n`` -- positive integer indicating the degree

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: l = lambda c: [ (i[0],[j for j in sorted(i[1].items())]) for i in sorted(c.items())]
            sage: JP._m_cache(2)
            sage: l(JP._self_to_m_cache[2])
            [([1, 1], [([1, 1], 1)]), ([2], [([1, 1], 2/(t + 1)), ([2], 1)])]
            sage: l(JP._m_to_self_cache[2])
            [([1, 1], [([1, 1], 1)]), ([2], [([1, 1], -2/(t + 1)), ([2], 1)])]
            sage: JP._m_cache(3)
            sage: l(JP._m_to_self_cache[3])
            [([1, 1, 1], [([1, 1, 1], 1)]),
             ([2, 1], [([1, 1, 1], -6/(t + 2)), ([2, 1], 1)]),
             ([3], [([1, 1, 1], 6/(t^2 + 3*t + 2)), ([2, 1], -3/2/(t + 1/2)), ([3], 1)])]
            sage: l(JP._self_to_m_cache[3])
            [([1, 1, 1], [([1, 1, 1], 1)]),
             ([2, 1], [([1, 1, 1], 6/(t + 2)), ([2, 1], 1)]),
             ([3],
              [([1, 1, 1], 3/(t^2 + 3/2*t + 1/2)), ([2, 1], 3/2/(t + 1/2)), ([3], 1)])]
        """
        if n in self._self_to_m_cache:
            return
        self._self_to_m_cache[n] = {}
        t = QQt.gen()
        monomial = sage.combinat.sf.sf.SymmetricFunctions(QQt).monomial()
        JP = sage.combinat.sf.sf.SymmetricFunctions(QQt).jack().P()
        JP._gram_schmidt(n, monomial, lambda p: part_scalar_jack(p, p, t),
                         self._self_to_m_cache[n], upper_triangular=True)
        JP._invert_morphism(n, QQt, self._self_to_m_cache,
                            self._m_to_self_cache, to_other_function=self._to_m)

    def _to_m(self, part):
        r"""
        Return a function that takes in a partition lambda that returns the
        coefficient of lambda in the expansion of self(part) in the
        monomial basis.

        This assumes that the cache from the Jack polynomials in the `P`
        basis to the monomial symmetric functions has already been
        computed.

        INPUT:

        - ``self`` -- an instance of the Jack `P` basis of the symmetric functions
        - ``part`` -- a partition

        OUTPUT:

        - returns a function that accepts a partition and returns the coefficients
          of the expansion of the element of ``P(part)`` in the monomial basis

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: JP._m_cache(3)
            sage: f = JP._to_m(Partition([2,1]))
            sage: [f(part) for part in Partitions(3)]
            [0, 1, 6/(t + 2)]
            sage: JP.symmetric_function_ring().m()(JP[2,1])
            (6/(t+2))*m[1, 1, 1] + m[2, 1]
        """
        f = lambda part2: self._self_to_m_cache[sum(part)][part].get(part2, 0)
        return f

    def product(self, left, right):
        r"""
        The product of two Jack symmetric functions is done by multiplying the
        elements in the monomial basis and then expressing the elements
        the basis ``self``.

        INPUT:

        - ``self`` -- a Jack basis of the symmetric functions
        - ``left``, ``right`` -- symmetric function elements

        OUTPUT: the product of ``left`` and ``right`` expanded in the basis ``self``

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: m = JP.symmetric_function_ring().m()
            sage: JP([1])^2 # indirect doctest
            (2*t/(t+1))*JackP[1, 1] + JackP[2]
            sage: m(_)
            2*m[1, 1] + m[2]
            sage: JP = SymmetricFunctions(QQ).jack(t=2).P()
            sage: JP([2,1])^2
            125/63*JackP[2, 2, 1, 1] + 25/12*JackP[2, 2, 2] + 25/18*JackP[3, 1, 1, 1] + 12/5*JackP[3, 2, 1] + 4/3*JackP[3, 3] + 4/3*JackP[4, 1, 1] + JackP[4, 2]
            sage: m(_)
            45*m[1, 1, 1, 1, 1, 1] + 51/2*m[2, 1, 1, 1, 1] + 29/2*m[2, 2, 1, 1] + 33/4*m[2, 2, 2] + 9*m[3, 1, 1, 1] + 5*m[3, 2, 1] + 2*m[3, 3] + 2*m[4, 1, 1] + m[4, 2]
        """
        return self(self._m(left) * self._m(right))

    def scalar_jack_basis(self, part1, part2=None):
        r"""
        Return the scalar product of `P(part1)` and `P(part2)`.

        This is equation (10.16) of [Mc1995]_ on page 380.

        INPUT:

        - ``self`` -- an instance of the Jack `P` basis of the symmetric functions
        - ``part1`` -- a partition
        - ``part2`` -- an optional partition (default: ``None``)

        OUTPUT:

        - the scalar product between `P(part1)` and `P(part2)` (or itself if `part2` is None)

        REFERENCES:

        .. [Mc1995] \I. G. Macdonald, Symmetric functions and Hall
           polynomials, second ed., The Clarendon Press, Oxford
           University Press, New York, 1995, With contributions by
           A. Zelevinsky, Oxford Science Publications.

        EXAMPLES::

            sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
            sage: JJ = SymmetricFunctions(FractionField(QQ['t'])).jack().J()
            sage: JP.scalar_jack_basis(Partition([2,1]), Partition([1,1,1]))
            0
            sage: JP._normalize_coefficients(JP.scalar_jack_basis(Partition([3,2,1]), Partition([3,2,1])))
            (6*t^6 + 10*t^5 + 11/2*t^4 + t^3)/(t^3 + 11/2*t^2 + 10*t + 6)
            sage: JJ(JP[3,2,1]).scalar_jack(JP[3,2,1])
            (6*t^6 + 10*t^5 + 11/2*t^4 + t^3)/(t^3 + 11/2*t^2 + 10*t + 6)

        With a single argument, takes `part2 = part1`::

            sage: JP.scalar_jack_basis(Partition([2,1]), Partition([2,1]))
            (2*t^3 + t^2)/(t + 2)
            sage: JJ(JP[2,1]).scalar_jack(JP[2,1])
            (2*t^3 + t^2)/(t + 2)
        """
        if part2 is not None and part1 != part2:
            return self.base_ring().zero()
        return self.c2(part1) / self.c1(part1)

    class Element(JackPolynomials_generic.Element):
        def scalar_jack(self, x, t=None):
            r"""
            The scalar product on the symmetric functions where the power sums
            are orthogonal and `\langle p_\mu, p_\mu \rangle = z_\mu t^{length(mu)}`
            where the t parameter from the Jack symmetric function family.

            INPUT:

            - ``self`` -- an element of the Jack `P` basis
            - ``x`` -- an element of the `P` basis

            EXAMPLES::

                sage: JP = SymmetricFunctions(FractionField(QQ['t'])).jack().P()
                sage: l = [JP(p) for p in Partitions(3)]
                sage: matrix([[a.scalar_jack(b) for a in l] for b in l])
                [3*t^3/(t^2 + 3/2*t + 1/2)                         0                         0]
                [                        0     (2*t^3 + t^2)/(t + 2)                         0]
                [                        0                         0 1/6*t^3 + 1/2*t^2 + 1/3*t]
            """
            if isinstance(x, JackPolynomials_p) and t is None:
                P = self.parent()
                return P._apply_multi_module_morphism(self, x, P.scalar_jack_basis, orthogonal=True)
            else:
                return JackPolynomials_generic.Element.scalar_jack(self, x, t)

#J basis


class JackPolynomials_j(JackPolynomials_generic):

    def __init__(self, jack):
        r"""
        The `J` basis is a defined as a normalized form of the `P` basis.

        INPUT:

        - ``self`` -- an instance of the Jack `P` basis of the symmetric functions
        - ``jack`` -- a family of Jack symmetric function bases

        EXAMPLES::

            sage: # needs lrcalc_python
            sage: J = SymmetricFunctions(FractionField(QQ['t'])).jack().J()
            sage: TestSuite(J).run(skip=['_test_associativity', '_test_distributivity', '_test_prod']) # products are too expensive
            sage: TestSuite(J).run(elements = [J.t*J[1,1]+J[2], J[1]+(1+J.t)*J[1,1]])  # long time (3s on sage.math, 2012)
        """
        self._name = "Jack polynomials in the J basis"
        self._prefix = "JackJ"
        JackPolynomials_generic.__init__(self, jack)

        # Should be shared with _q (and possibly other bases in Macdo/HL) as BasesByRenormalization
        self._P = self._jack.P()
        # temporary until Hom(GradedHopfAlgebrasWithBasis) works better
        category = sage.categories.all.ModulesWithBasis(self.base_ring())
        phi = self.module_morphism(diagonal=self.c1,
                                   codomain=self._P, category=category)
        # should use module_morphism(on_coeffs = ...) once it exists
        self._P.register_coercion(self._P._normalize_morphism(category) * phi)
        self   .register_coercion(self   ._normalize_morphism(category) * ~phi)

    class Element(JackPolynomials_generic.Element):
        pass


#Q basis
class JackPolynomials_q(JackPolynomials_generic):

    def __init__(self, jack):
        r"""
        The `Q` basis is defined as a normalized form of the `P` basis.

        INPUT:

        - ``self`` -- an instance of the Jack `Q` basis of the symmetric functions
        - ``jack`` -- a family of Jack symmetric function bases

        EXAMPLES::

            sage: # needs lrcalc_python
            sage: Q = SymmetricFunctions(FractionField(QQ['t'])).jack().Q()
            sage: TestSuite(Q).run(skip=['_test_associativity', '_test_distributivity', '_test_prod']) # products are too expensive
            sage: TestSuite(Q).run(elements = [Q.t*Q[1,1]+Q[2], Q[1]+(1+Q.t)*Q[1,1]])  # long time (3s on sage.math, 2012)
        """
        self._name = "Jack polynomials in the Q basis"
        self._prefix = "JackQ"
        JackPolynomials_generic.__init__(self, jack)

        # Should be shared with _j (and possibly other bases in Macdo/HL) as BasesByRenormalization
        self._P = self._jack.P()
        # temporary until Hom(GradedHopfAlgebrasWithBasis) works better
        category = sage.categories.all.ModulesWithBasis(self.base_ring())
        phi = self._P.module_morphism(diagonal=self._P.scalar_jack_basis,
                                      codomain=self, category=category)
        self.register_coercion(self._normalize_morphism(category) * phi)
        self._P.register_coercion(self._P._normalize_morphism(category) * ~phi)

    class Element(JackPolynomials_generic.Element):
        pass


qp_to_h_cache = {}
h_to_qp_cache = {}


class JackPolynomials_qp(JackPolynomials_generic):
    def __init__(self, jack):
        r"""
        The `Qp` basis is the dual basis to the `P` basis with respect to the
        standard scalar product

        INPUT:

        - ``self`` -- an instance of the Jack `Qp` basis of the symmetric functions
        - ``jack`` -- a family of Jack symmetric function bases

        EXAMPLES::

            sage: Qp = SymmetricFunctions(FractionField(QQ['t'])).jack().Qp()
            sage: TestSuite(Qp).run(skip=['_test_associativity', '_test_distributivity', '_test_prod']) # products are too expensive
            sage: TestSuite(Qp).run(elements = [Qp.t*Qp[1,1]+Qp[2], Qp[1]+(1+Qp.t)*Qp[1,1]])  # long time (3s on sage.math, 2012)
        """
        self._name = "Jack polynomials in the Qp basis"
        self._prefix = "JackQp"
        JackPolynomials_generic.__init__(self, jack)
        self._P = self._jack.P()
        self._self_to_h_cache = qp_to_h_cache
        self._h_to_self_cache = h_to_qp_cache

    def product(self, left, right):
        r"""
        The product of two Jack symmetric functions is done by multiplying the
        elements in the monomial basis and then expressing the elements
        the basis ``self``.

        INPUT:

        - ``self`` -- an instance of the Jack `Qp` basis of the symmetric functions
        - ``left``, ``right`` -- symmetric function elements

        OUTPUT: the product of ``left`` and ``right`` expanded in the basis ``self``

        EXAMPLES::

            sage: JQp = SymmetricFunctions(FractionField(QQ['t'])).jack().Qp()
            sage: h = JQp.symmetric_function_ring().h()
            sage: JQp([1])^2 # indirect doctest
            JackQp[1, 1] + (2/(t+1))*JackQp[2]
            sage: h(_)
            h[1, 1]
            sage: JQp = SymmetricFunctions(QQ).jack(t=2).Qp()
            sage: h = SymmetricFunctions(QQ).h()
            sage: JQp([2,1])^2
            JackQp[2, 2, 1, 1] + 2/3*JackQp[2, 2, 2] + 2/3*JackQp[3, 1, 1, 1] + 48/35*JackQp[3, 2, 1] + 28/75*JackQp[3, 3] + 128/225*JackQp[4, 1, 1] + 28/75*JackQp[4, 2]
            sage: h(_)
            h[2, 2, 1, 1] - 6/5*h[3, 2, 1] + 9/25*h[3, 3]
        """
        return self(self._h(left) * self._h(right))

    def _h_cache(self, n):
        r"""
        Compute the change of basis between the Jack polynomials in the `Qp`
        basis and the homogeneous symmetric functions. This uses the coefficients
        in the change of basis between the Jack `P` basis and the monomial basis.

        INPUT:

        - ``self`` -- an instance of the Jack `Qp` basis of the symmetric functions
        - ``n`` -- positive integer indicating the degree

        EXAMPLES::

            sage: JQp = SymmetricFunctions(FractionField(QQ['t'])).jack().Qp()
            sage: l = lambda c: [ (i[0],[j for j in sorted(i[1].items())]) for i in sorted(c.items())]
            sage: JQp._h_cache(2)
            sage: l(JQp._self_to_h_cache[2])
            [([1, 1], [([1, 1], 1), ([2], -2/(t + 1))]), ([2], [([2], 1)])]
            sage: l(JQp._h_to_self_cache[2])
            [([1, 1], [([1, 1], 1), ([2], 2/(t + 1))]), ([2], [([2], 1)])]
            sage: JQp._h_cache(3)
            sage: l(JQp._h_to_self_cache[3])
            [([1, 1, 1],
              [([1, 1, 1], 1), ([2, 1], 6/(t + 2)), ([3], 3/(t^2 + 3/2*t + 1/2))]),
             ([2, 1], [([2, 1], 1), ([3], 3/2/(t + 1/2))]),
             ([3], [([3], 1)])]
            sage: l(JQp._self_to_h_cache[3])
            [([1, 1, 1], [([1, 1, 1], 1), ([2, 1], -6/(t + 2)), ([3], 6/(t^2 + 3*t + 2))]),
             ([2, 1], [([2, 1], 1), ([3], -3/2/(t + 1/2))]),
             ([3], [([3], 1)])]
        """
        if n in self._self_to_h_cache:
            return
        else:
            self._self_to_h_cache[n] = {}
            self._h_to_self_cache[n] = {}
        self._P._m_cache(n)
        from_cache_1 = self._P._self_to_m_cache[n]
        to_cache_1 = self._self_to_h_cache[n]
        from_cache_2 = self._P._m_to_self_cache[n]
        to_cache_2 = self._h_to_self_cache[n]
        for mu in from_cache_1:
            for la in from_cache_1[mu]:
                if la not in to_cache_1:
                    to_cache_1[la] = {}
                    to_cache_2[la] = {}
                to_cache_2[la][mu] = from_cache_1[mu][la]
                to_cache_1[la][mu] = from_cache_2[mu][la]

    def _self_to_h( self, x ):
        r"""
        Isomorphism from ``self`` to the homogeneous basis.

        INPUT:

        - ``self`` -- a Jack `Qp` basis of the symmetric functions
        - ``x`` -- an element of the Jack `Qp` basis

        OUTPUT: an element of the homogeneous basis equivalent to ``x``

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JQp = Sym.jack(t=2).Qp()
            sage: h = Sym.homogeneous()
            sage: JQp._self_to_h(JQp[2,1])
            h[2, 1] - 3/5*h[3]

        This is for internal use only. Please use instead::

            sage: h(JQp[2,1])
            h[2, 1] - 3/5*h[3]
        """
        return self._h._from_cache(x, self._h_cache, self._self_to_h_cache,
                                   t=self.t)

    def _h_to_self(self, x):
        r"""
        Isomorphism from the homogeneous basis into ``self``.

        INPUT:

        - ``self`` -- a Jack `Qp` basis of the symmetric functions
        - ``x`` -- element of the homogeneous basis

        OUTPUT: an element of the Jack `Qp` basis equivalent to ``x``

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ)
            sage: JQp = Sym.jack(t=2).Qp()
            sage: h = Sym.homogeneous()
            sage: JQp._h_to_self(h[2,1])
            JackQp[2, 1] + 3/5*JackQp[3]

        This is for internal use only. Please use instead::

            sage: JQp(h[2,1])
            JackQp[2, 1] + 3/5*JackQp[3]
        """
        return self._from_cache(x, self._h_cache, self._h_to_self_cache,
                                t=self.t)

    def coproduct_by_coercion(self, elt):
        r"""
        Return the coproduct of the element ``elt`` by coercion to the Schur basis.

        INPUT:

        - ``elt`` -- an instance of the ``Qp`` basis

        OUTPUT:

        - The coproduct acting on ``elt``, the result is an element of the
          tensor squared of the ``Qp`` symmetric function basis

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ['t'].fraction_field())
            sage: JQp = Sym.jack().Qp()
            sage: JQp[2,2].coproduct()   #indirect doctest
            JackQp[] # JackQp[2, 2] + (2*t/(t+1))*JackQp[1] # JackQp[2, 1] + JackQp[1, 1] # JackQp[1, 1] + ((2*t^3+4*t^2)/(t^3+5/2*t^2+2*t+1/2))*JackQp[2] # JackQp[2] + (2*t/(t+1))*JackQp[2, 1] # JackQp[1] + JackQp[2, 2] # JackQp[]
        """
        h = elt.parent().realization_of().h()
        parent = elt.parent()
        from sage.categories.tensor import tensor
        cfunc = lambda x, y: tensor([parent(x), parent(y)])
        cprod = h(elt).coproduct().apply_multilinear_morphism( cfunc )
        normalize = lambda c: normalize_coefficients( parent, c )
        return cprod.parent().sum(normalize(coeff)*tensor([parent(x), parent(y)])
                        for ((x,y), coeff) in cprod)

    class Element(JackPolynomials_generic.Element):
        pass

#Zonal polynomials ( =P(at t=2) )


class SymmetricFunctionAlgebra_zonal(sfa.SymmetricFunctionAlgebra_generic):
    def __init__(self, Sym):
        r"""
        Return the algebra of zonal polynomials.

        INPUT:

        - ``self`` -- a zonal basis of the symmetric functions
        - ``Sym`` -- a ring of the symmetric functions

        EXAMPLES::

            sage: # needs lrcalc_python
            sage: Z = SymmetricFunctions(QQ).zonal()
            sage: Z([2])^2
            64/45*Z[2, 2] + 16/21*Z[3, 1] + Z[4]
            sage: Z = SymmetricFunctions(QQ).zonal()
            sage: TestSuite(Z).run(skip=['_test_associativity', '_test_distributivity', '_test_prod']) # products are too expensive
            sage: TestSuite(Z).run(elements = [Z[1,1]+Z[2], Z[1]+2*Z[1,1]])
        """
        self._sym = Sym
        self._jack = self._sym.jack(t=2)
        self._P = self._jack.P()
        #self._m_to_self_cache = {} Now that we compute Jacks once, there is a global cache
        #self._self_to_m_cache = {} and we don't need to compute it separately for zonals
        sfa.SymmetricFunctionAlgebra_generic.__init__(self, self._sym,
                                                      prefix='Z', basis_name='zonal')
        category = sage.categories.all.ModulesWithBasis(self._sym.base_ring())
        self   .register_coercion(SetMorphism(Hom(self._P, self, category), self.sum_of_terms))
        self._P.register_coercion(SetMorphism(Hom(self, self._P, category), self._P.sum_of_terms))

    def product(self, left, right):
        r"""
        The product of two zonal symmetric functions is done by multiplying the
        elements in the monomial basis and then expressing the elements
        in the basis ``self``.

        INPUT:

        - ``self`` -- a zonal basis of the symmetric functions
        - ``left``, ``right`` -- symmetric function elements

        OUTPUT: the product of ``left`` and ``right`` expanded in the basis ``self``

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ)
            sage: Z = Sym.zonal()
            sage: JP = Sym.jack(t=1).P()
            sage: Z([2])*Z([3])                    # indirect doctest
            192/175*Z[3, 2] + 32/45*Z[4, 1] + Z[5]
            sage: Z([2])*JP([2])
            10/27*Z[2, 1, 1] + 64/45*Z[2, 2] + 23/21*Z[3, 1] + Z[4]
            sage: JP = Sym.jack(t=2).P()
            sage: Z([2])*JP([2])
            64/45*Z[2, 2] + 16/21*Z[3, 1] + Z[4]
        """
        return self(self._P(left) * self._P(right))

    class Element(sfa.SymmetricFunctionAlgebra_generic.Element):
        def scalar_zonal(self, x):
            r"""
            The zonal scalar product has the power sum basis and the zonal
            symmetric functions are orthogonal. In particular,
            `\langle p_\mu, p_\mu \rangle = z_\mu 2^{length(\mu)}`.

            INPUT:

            - ``self`` -- an element of the zonal basis
            - ``x`` -- an element of the symmetric function

            OUTPUT: the scalar product between ``self`` and ``x``

            EXAMPLES::

                sage: Sym = SymmetricFunctions(QQ)
                sage: Z = Sym.zonal()
                sage: parts = Partitions(3).list()
                sage: matrix([[Z(a).scalar_zonal(Z(b)) for a in parts] for b in parts])
                [16/5    0    0]
                [   0    5    0]
                [   0    0    4]
                sage: p = Z.symmetric_function_ring().power()
                sage: matrix([[Z(p(a)).scalar_zonal(p(b)) for a in parts] for b in parts])
                [ 6  0  0]
                [ 0  8  0]
                [ 0  0 48]
            """
            P = self.parent()._P
            return P(self).scalar_jack(P(x),2)


# Backward compatibility for unpickling
from sage.misc.persist import register_unpickle_override

register_unpickle_override('sage.combinat.sf.jack', 'JackPolynomial_qp', JackPolynomials_qp.Element)
register_unpickle_override('sage.combinat.sf.jack', 'JackPolynomial_j', JackPolynomials_j.Element)
register_unpickle_override('sage.combinat.sf.jack', 'JackPolynomial_p', JackPolynomials_p.Element)
register_unpickle_override('sage.combinat.sf.jack', 'JackPolynomial_q', JackPolynomials_q.Element)
#register_unpickle_override('sage.combinat.sf.jack', 'SymmetricFunctionAlgebra_zonal',  SymmetricFunctionAlgebra_zonal.Element)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Bijection classes for type `C_n^{(1)}`

Part of the (internal) classes which runs the bijection between rigged
configurations and KR tableaux of type `C_n^{(1)}`.

AUTHORS:

- Travis Scrimshaw (2012-12-21): Initial version

TESTS::

    sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['C', 3, 1], [[2,1]])
    sage: from sage.combinat.rigged_configurations.bij_type_C import KRTToRCBijectionTypeC
    sage: bijection = KRTToRCBijectionTypeC(KRT(pathlist=[[-1,2]]))
    sage: TestSuite(bijection).run()
    sage: RC = RiggedConfigurations(['C', 3, 1], [[2, 1]])
    sage: from sage.combinat.rigged_configurations.bij_type_C import RCToKRTBijectionTypeC
    sage: bijection = RCToKRTBijectionTypeC(RC(partition_list=[[],[],[]]))
    sage: TestSuite(bijection).run()
"""

# ****************************************************************************
#       Copyright (C) 2012 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.rigged_configurations.bij_type_A import KRTToRCBijectionTypeA
from sage.combinat.rigged_configurations.bij_type_A import RCToKRTBijectionTypeA


class KRTToRCBijectionTypeC(KRTToRCBijectionTypeA):
    r"""
    Specific implementation of the bijection from KR tableaux to rigged
    configurations for type `C_n^{(1)}`.

    This inherits from type `A_n^{(1)}` because we use the same methods in
    some places.
    """

    def next_state(self, val):
        r"""
        Build the next state for type `C_n^{(1)}`.

        TESTS::

            sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['C', 3, 1], [[2,1]])
            sage: from sage.combinat.rigged_configurations.bij_type_C import KRTToRCBijectionTypeC
            sage: bijection = KRTToRCBijectionTypeC(KRT(pathlist=[[-1,2]]))
            sage: bijection.cur_path.insert(0, [])
            sage: bijection.cur_dims.insert(0, [0, 1])
            sage: bijection.cur_path[0].insert(0, [2])
            sage: bijection.next_state(2)
        """
        # Note that we must subtract 1 from n to match the indices.
        n = self.n
        tableau_height = len(self.cur_path[0]) - 1

        # If it is a regular value, we follow the A_n rules
        if val > 0:
            KRTToRCBijectionTypeA.next_state(self, val)
            return

        pos_val = -val
        case_S = [None] * n

        # Always add a cell to the first singular value in the first
        #   tableau we are updating.
        if len(self.ret_rig_con[pos_val - 1]) > 0:
            max_width = self.ret_rig_con[pos_val - 1][0]
        else:
            max_width = 1

        # Special case for inserting -n
        if pos_val == n:
            max_width *= 2

        # Add cells similar to type A_n but we move to the right until we
        #   reach the value of n
        for a in range(pos_val - 1, n - 1):
            max_width = self.ret_rig_con[a].insert_cell(max_width)
            case_S[a] = max_width

        # Special case for n
        max_width = self.ret_rig_con[n-1].insert_cell(max_width // 2) * 2

        # Now go back following the special C_n rules
        for a in reversed(range(tableau_height, n - 1)):
            if case_S[a] == max_width:
                self._insert_cell_case_S(self.ret_rig_con[a])
            else:
                max_width = self.ret_rig_con[a].insert_cell(max_width)
            self._update_vacancy_nums(a + 1)
            self._update_partition_values(a + 1)

        # Update the final rigged partitions
        if tableau_height < n:
            self._update_vacancy_nums(tableau_height)
            self._update_partition_values(tableau_height)

        if pos_val <= tableau_height:
            for a in range(pos_val-1, tableau_height):
                self._update_vacancy_nums(a)
                self._update_partition_values(a)
            if pos_val > 1:
                self._update_vacancy_nums(pos_val - 2)
                self._update_partition_values(pos_val - 2)
        elif tableau_height > 0:
            self._update_vacancy_nums(tableau_height - 1)
            self._update_partition_values(tableau_height - 1)

    def _insert_cell_case_S(self, partition):
        """
        Insert a cell when case `(S)` holds.

        TESTS::

            sage: RC = RiggedConfigurations(['C', 2, 1], [[2, 2]])
            sage: RP = RC(partition_list=[[2],[2,2]])[1]
            sage: RP
            -4[ ][ ]-4
            -4[ ][ ]-4
            <BLANKLINE>
            sage: RP.rigging[0] = None
            sage: from sage.combinat.rigged_configurations.bij_type_C import KRTToRCBijectionTypeC
            sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(['C', 3, 1], [[2,1]])
            sage: bijection = KRTToRCBijectionTypeC(KRT(pathlist=[[-1,2]]))
            sage: bijection._insert_cell_case_S(RP)
            sage: RP
            -4[ ][ ][ ]None
            -4[ ][ ]-4
            <BLANKLINE>
        """
        # Special case when adding twice to the first row
        if partition.rigging[0] is None:
            partition._list[0] += 1
            return

        num_rows = len(partition)
        for i in reversed(range(1, num_rows)):
            if partition.rigging[i] is None:
                j = i - 1
                while j >= 0 and partition._list[j] == partition._list[i]:
                    partition.rigging[j+1] = partition.rigging[j]  # Shuffle it along
                    j -= 1
                partition._list[j+1] += 1
                partition.rigging[j+1] = None
                return


class RCToKRTBijectionTypeC(RCToKRTBijectionTypeA):
    r"""
    Specific implementation of the bijection from rigged configurations to
    tensor products of KR tableaux for type `C_n^{(1)}`.
    """

    def next_state(self, height):
        r"""
        Build the next state for type `C_n^{(1)}`.

        TESTS::

            sage: RC = RiggedConfigurations(['C', 3, 1], [[2, 1]])
            sage: from sage.combinat.rigged_configurations.bij_type_C import RCToKRTBijectionTypeC
            sage: bijection = RCToKRTBijectionTypeC(RC(partition_list=[[2],[2],[1]]))
            sage: bijection.next_state(1)
            -1
        """
        height -= 1  # indexing
        n = self.n
        ell = [None] * (2*n)
        case_S = [False] * n
        b = None

        # Calculate the rank and ell values

        last_size = 0
        for a in range(height, n-1):
            ell[a] = self._find_singular_string(self.cur_partitions[a], last_size)

            if ell[a] is None:
                b = a + 1
                break
            else:
                last_size = self.cur_partitions[a][ell[a]]

        # Special case for n
        if b is None:
            # Since we are dividing by 2, we can use the identity of
            #   ceiling = floor + remainder
            ell[n-1] = self._find_singular_string(self.cur_partitions[n-1],
                                                 (last_size // 2) + (last_size % 2))

            if ell[n-1] is None:
                b = n
            else:
                last_size = self.cur_partitions[n-1][ell[n-1]] * 2

        if b is None:
            # Now go back
            ell[2*n-1] = ell[n-1]
            case_S[n-1] = True
            for a in reversed(range(n-1)):
                if a >= height and self.cur_partitions[a][ell[a]] == last_size:
                    ell[n+a] = ell[a]
                    case_S[a] = True
                else:  # note last_size > 1
                    ell[n+a] = self._find_singular_string(self.cur_partitions[a], last_size)

                    if ell[n + a] is None:
                        b = -(a + 2)
                        break
                    else:
                        last_size = self.cur_partitions[a][ell[n + a]]

        if b is None:
            b = -1

        # Determine the new rigged configuration by removing boxes from the
        #   selected string and then making the new string singular
        if case_S[0]:
            row_num = self.cur_partitions[0].remove_cell(ell[0], 2)
            row_num_bar = None
        else:
            row_num = self.cur_partitions[0].remove_cell(ell[0])
            row_num_bar = self.cur_partitions[0].remove_cell(ell[n])
        for a in range(1, n-1):
            if case_S[a]:
                row_num_next = self.cur_partitions[a].remove_cell(ell[a], 2)
                row_num_bar_next = None
            else:
                row_num_next = self.cur_partitions[a].remove_cell(ell[a])
                row_num_bar_next = self.cur_partitions[a].remove_cell(ell[n+a])

            self._update_vacancy_numbers(a - 1)
            if row_num is not None:
                self.cur_partitions[a-1].rigging[row_num] = self.cur_partitions[a-1].vacancy_numbers[row_num]
            if row_num_bar is not None:
                self.cur_partitions[a-1].rigging[row_num_bar] = self.cur_partitions[a-1].vacancy_numbers[row_num_bar]
            row_num = row_num_next
            row_num_bar = row_num_bar_next

        row_num_next = self.cur_partitions[n-1].remove_cell(ell[n-1])

        self._update_vacancy_numbers(n - 2)
        if row_num is not None:
            self.cur_partitions[n-2].rigging[row_num] = self.cur_partitions[n-2].vacancy_numbers[row_num]
        if row_num_bar is not None:
            self.cur_partitions[n-2].rigging[row_num_bar] = self.cur_partitions[n-2].vacancy_numbers[row_num_bar]

        self._update_vacancy_numbers(n - 1)
        if row_num_next is not None:
            self.cur_partitions[n-1].rigging[row_num_next] = self.cur_partitions[n-1].vacancy_numbers[row_num_next]

        return b
